function app = rbflindisc(model, cfg)
% Create RBF-linear state approximator with discrete actions

CFG.type = 'rbflindisc';            % # of samples to generate
CFG.disc = 1;
CFG.c = [];
CFG.rad = [];
CFG.ugrids = [];

cfg = parseconfig(cfg, CFG);

if isempty(cfg.c) || isempty(cfg.rad) || isempty(cfg.ugrids),
    error('Centers, radii of RBFs, or discrete actions, could not be obtained');
end;

app.type = cfg.type;
% centers & radii
app.c = cfg.c;
app.rad = cfg.rad;
app.Nc = size(app.c, 2);
if (size(app.c, 1) ~= model.p) || any(size(app.c) ~= size(app.rad)),
    error('Incorrect c or rad size');
end;
% action space, structured and flat
app.U = cfg.ugrids; app.Uflat = flat(cfg.ugrids);
% number of actions
app.M = size(app.Uflat, 2);
app.Nt = model.p + 1;       % number of terms (linear + bias)
app.Nx = app.Nc * app.Nt;   % total number of state basis functions
app.N = app.Nx * app.M;     % total number of basis functions

app.phi = @rbflindisc_phi; app.q = @rbflindisc_q; app.h = @rbflindisc_h; 
% publish also index-phi function, so that algorithms that support index
% optimization can use it
app.indphi = @rbflindisc_indphi;
app.Nindphi = app.Nx;

% plot functions
app.plotv = @rbflindisc_plotv; app.ploth = @rbflindisc_ploth;

end     % TRIANG constructor

% ===== Core functions
% BF values in x, u
function phi = rbflindisc_phi(app, x, u)
phi = zeros(app.N, 1);
ui = findflat(u, app.Uflat, 1, 'first');
rbfphi = nrbf(x, app.Nc, app.c, app.rad)';
xm = repmat(x', app.Nc, 1);
phi( ((ui-1)*app.Nx)+1 : (ui*app.Nx) ) = [repmat(rbfphi, app.Nt-1, 1) .* xm(:); rbfphi];
end
% Index-BF values in x, u (for index-optimized implementations)
function [offset, phi] = rbflindisc_indphi(app, x, u)
offset = (findflat(u, app.Uflat, 1, 'first')-1) * app.Nx;
rbfphi = nrbf(x, app.Nc, app.c, app.rad)';
xm = repmat(x', app.Nc, 1);
phi = [repmat(rbfphi, app.Nt-1, 1) .* xm(:); rbfphi];
end
% Q-value in x, u
function q = rbflindisc_q(app, theta, x, u)
ui = findflat(u, app.Uflat, 1, 'first');
rbfphi = nrbf(x, app.Nc, app.c, app.rad);
xm = repmat(x', app.Nc, 1);
q = [repmat(rbfphi, 1, app.Nt-1) .* xm(:)', rbfphi] * theta( ((ui-1)*app.Nx)+1 : (ui*app.Nx) );
end
% Greedy policy in x
function u = rbflindisc_h(app, theta, x)
rbfphi = nrbf(x, app.Nc, app.c, app.rad);
xm = repmat(x', app.Nc, 1);
qx = [repmat(rbfphi, 1, app.Nt-1) .* xm(:)', rbfphi] * reshape(theta, app.Nx, app.M);
u = app.Uflat(:, find(qx == max(qx), 1, 'first'));
end


% ----- Plot functions
% Plot V - forward call to generic plot fun
function rbflindisc_plotv(r, theta, varargin)
for i = size(r.c, 1):-1:1, xb(i, 1:2) = [min(r.c(i, :)) max(r.c(i, :))]; end;
for i = length(r.U):-1:1, ub(i, 1:2) = r.U{i}([1 end]); end;
approx_plotv(r, theta, xb, ub, varargin{:});
end
% Plot h - forward call to generic plot fun
function rbflindisc_ploth(r, theta, varargin)
for i = size(r.c, 1):-1:1, xb(i, 1:2) = [min(r.c(i, :)) max(r.c(i, :))]; end;
for i = length(r.U):-1:1, ub(i, 1:2) = r.U{i}([1 end]); end;
approx_ploth(r, theta, xb, ub, varargin{:});
end
