function varargout = elev(mode, varargin)
% Elevator function
%   MODEL = ELEV('model', CFG)
%   X0 = ELEV('reset', MODEL)
%   [XPLUS, RPLUS, TERMINAL] = ELEV('step', MODEL, X, U, K)
%   ELEV('visualizestep', MODEL, X, U, XPLUS, RPLUS, K)
%
% Inputs:
%   MODE    - selector for mode in which function will run
%   CFG     - configuration for model creation
%   MODEL   - elevator model (structure)
%   X       - current state
%   U       - current action
%   XPLUS   - next state
%   RPLUS   - received reward
%   K       - current time step
%
% Outputs:
%   MODEL   - created model
%   X0      - initial state
%   XPLUS   - next state
%   RPLUS   - received reward
%   TERMINAL- whether XPLUS is terminal

switch mode,
    case 'model',
        cfg = varargin{1};
        % default configuration values
        % episodic=1: terminate trial when elevator stops at ground floor, and there are no calls
        % episodic=0: keep running "forever"
        CFG.episodic = 0;
        % building/elevator properties
        CFG.h = 4;              % building height (not including 0th, ground floor)
        CFG.cap = 3;            % capacity of elevator
        CFG.maxcalls = 3;       % maximum calls at each floor (any other arriving calls are ignored)
        % call mode: 'rand' for uniform random, 'fixed' for given in advance, 'none' for no calls
        % after initial state is set
        CFG.callmode = 'rand';
        % call probability/ies, for 'rand' callmode
        % if single number, calls are equiprobable at each nonzero floor
        % if h-element column vector, different call probabilities at each nonzero floor
        CFG.callprob = .01;     % calls arrive with 1% probability at each floor, at each step
        CFG.calltab = [];       % pre-given call table, if given in advance; h x maxsteps, 0 or 1
        %
        % how to reset in the beginning of trial
        % 'rand' for fully random, 'fixed' for fixed initial state
        CFG.resetmode = 'rand';
        CFG.x0 = [];            % reset state
        %
        % reward function config
        % 'deliver' = 1 for each delivered passenger
        % 'wait' = -# of waiting passengers at each step (inside or outside elevator)
        % (add more here)
        CFG.rewtype = 'deliver';
        %
        % Visualization config
        CFG.vis_pause = 0;      % -1 = special value, meaning: wait for keypress
        
        
        % install default values for fields not specified on the input config
        fn = fieldnames(CFG); 
        if isempty(cfg), cfg = struct; end;
        for i = 1:length(fn), if ~isfield(cfg, fn{i}), cfg.(fn{i}) = CFG.(fn{i}); end; end;
        
        % all configuration fields get transferred to the model structure
        model = cfg; clear cfg;
        
        % declare model function
        model.fun = @elev;
        
        % initialize state and action spaces
        % first state: position of elevator, from 0 to h
        model.X = {0: model.h};
        % next h states: call flags
        for i = 1:model.h, model.X{end+1} = 0:model.maxcalls; end;
        % occupancy state
        model.X{end+1} = 0:model.cap;
        model.U = {[-1 0 1]};   % down, stop, up
        
        model.p = length(model.X); model.q = length(model.U);
        model.dimx = zeros(model.p);
        for i = 1:model.p, model.dimx(i) = length(model.X{i}); end;
        
        varargout = {model};
        
    case 'reset',
        model = varargin{1};
        switch model.resetmode,
            case 'rand',
                x0ind = unidrnd(model.dimx);
                x0 = zeros(model.p, model.q);
                for i = 1:model.p, x0(i) = model.X{i}(x0ind(i)); end;
            case 'fixed',
                x0 = model.x0;
            otherwise,
                error('elev: resetmode [%s] unsupported', model.resetmode);
        end;
        varargout = {x0};
        
    case 'step', 
        model = varargin{1}; x = varargin{2}; u = varargin{3}; k = varargin{4};
        % pick up individual state variables; by default they stay the same at next step
        pos = x(1); posplus = pos;
        calls = x(2+(0:model.h-1)); callsplus = calls;
        occ = x(end); occplus = occ;
        
        terminal = 0;
        % a) elevator motion and passenger pickup/deliver
        if u == 0,          % stopped, position stays the same, handle passenger transfers
            if pos == 0,
                delivered = occ;
                occplus = 0;
            else
                gettingin = min(model.cap - occ, calls(pos));
                occplus = occ + gettingin;
                callsplus(pos) = calls(pos) - gettingin;
            end;
        else                % just change position, no effect on calls or occupancy
            posplus = max(0, min(model.h, pos + u));
        end;
        
        % b) new calls
        switch model.callmode,
            case 'rand',
                newcalls = rand(model.h, 1) <= model.callprob;
            case 'fixed',
                % if we ran out of call settings, no calls
                if k <= size(model.callstab, 2),    newcalls = model.callstab(:, k);
                else                                newcalls = zeros(model.h, 1);
                end;
            case 'none',  
                newcalls = zeros(model.h, 1);
            otherwise,
                error('elev: callmode [%s] unsupported', model.callmode);
        end;
        % Note passengers arriving cannot "jump" directly into the elevator;
        % this could be changed
        callsplus = min(callsplus + newcalls, model.maxcalls);  % saturate at maxcalls

        % c) termination: if configured, and whenever we delivered and there are no new calls
        if model.episodic && pos == 0 && u == 0 && ~any(callsplus), terminal = 1; end;

        % d) compute reward
        switch model.rewtype,
            case 'deliver',
                if pos == 0 && u == 0,  rplus = delivered;
                else                    rplus = 0;
                end;
            case 'wait',
                rplus = -sum(callsplus) - occplus;
            otherwise,
                error('elev: rewtype [%s] unsupported', model.rewtype);
        end;
        
        xplus = [posplus; callsplus; occplus];
        varargout = {xplus, rplus, terminal};
        
    case 'visualizestep',
        model = varargin{1}; 
        x = varargin{2}; u = varargin{3}; xplus = varargin{4}; rplus = varargin{5}; k = varargin{6};
        pos = x(1); posplus = xplus(1);
        calls = x(2+ (0:model.h-1)); callsplus = xplus(2+ (0:model.h-1))    ;
        occ = x(end); occplus = xplus(end);
        
        % init spaces-filled representation
        v = char(' ' + zeros(4, 5*(model.h+1) + 1));      
        % add stuff to it
        % calls
        for i = 1:model.h,
            if callsplus(i) > 0, v(1, 5*i + 1 + (1:3)) = sprintf('C%2d', callsplus(i)); end;
        end;
        % elevator shaft
        v(3, :) = [repmat('|----', 1, model.h+1) '|'];
        % elevator & its occupancy
        v(3, 5*posplus + 1 + (1:4)) = sprintf('[%2d]', occplus);
        % elevator doors
        if u == 0,  v(2, 5*posplus + 1 + (1:4)) = '_  _';       % open
        else        v(2, 5*posplus + 1 + (1:4)) = '____';       % closed
        end;
        % action
        if u == -1,     v(3, 5*pos+1) = '<';            % moved left
        elseif u == 1,  v(3, 5*pos+6) = '>';            % moved right
        else    % u == 0
            if pos == 0,                     
                if occ > 0  % delivered passengers
                    v(2, 5*pos+(3:4)) = '^^';
                    v(1, 5*pos+(2:5)) = sprintf('^%2d^', occ);% ^ ^ = they are going on their merry way...
                end;
            elseif calls(pos) > 0 && occ < occplus,     % picked up passengers
                v(2, 5*pos+(3:4)) = 'vv';
            end;
        end;
        % status message
        s = sprintf('Elev, k=%d, r=%g', k, rplus);
        v(4, 1:length(s)) = s;
        
        % clear appropriate # of characters so that we overwrite older state
        if k > 1, disp(char(8 + zeros(1, numel(v) + 5))); end;
        % output the representation
        disp(v);
        if model.vis_pause == -1,   pause;
        elseif model.vis_pause > 0, pause(model.vis_pause); 
        end;
end;

end         % elev