function [evol, times, figh] = odetest(problem, cfg, setups)
% Benchmark for numerical integration methods
%   [EVOL, TIMES, FIGH] = ODETEST(PROBLEM, SETUPS, CFG)
%
% Parameters:
%   PROBLEM     - the problem function (preferably a function handle)
%   CFG         - configuration, see commented defaults. A control policy can be
%       specified as a configuration field.
%   SETUPS      - a structure containing the integration configurations to
%       test. One model obtained with 
%       PROBLEM('model', CFG.PREARGS{:}, SETUPS.FIELD{:}, CFG.POSTARGS{:}) 
%       for each FIELD will be used for integration. The PREARGS and POSTARGS are common
%       throughout the experiments.
% Returns:
%   EVOL        - a structure containing the trajectory history for every
%       numerical integration setup. A field EVOL.FIELD will be created for
%       every field FIELD in SETUPS, and will contain the trajectory
%       integrated with that numerical integration setup.
%   TIMES       - a structure containing the numerical integration runtime for every
%       numerical integration setup. A field TIMES.FIELD will be created for
%       every field FIELD in SETUPS, and will contain the runtime of the trajectory
%       integrated with that numerical integration setup.
%   FIGH        - handle of the figure containing the plotted trajectories

% Note uses old-style policy (no support for dynamic controllers)

% Default solvers: Euler, fixed-step RK, variable-step RK, stiff
% SETUPS.euler = 'odemethod=eul';
SETUPS.ode4N10 = 'odemethod=fixedode odesolver=@ode4 odesteps=10';
SETUPS.ode45 = 'odemethod=varode odesolver=@ode45';
% SETUPS.ode15s = 'odemethod=varode odesolver=@ode15s';
SETUPS.ode4N1 = 'odemethod=fixedode odesolver=@ode4 odesteps=1';
% SETUPS.ode4N5 = 'odemethod=fixedode odesolver=@ode4 odesteps=5';
% Default config
CFG.tend = 100;             % max length of simulation in time units (not steps)
CFG.x0 = [];                % initial state; problem default or 0 if not specified
CFG.policy = '';            % policy for control
CFG.policyargs = {};        % its arguments besides x and time
CFG.control = 1;            % 0 to suppress control plots; by default displays only if policy specified
CFG.plotfuns = {};          % variable-wise plot functions
CFG.legfontsize = 9;        % legend font size
CFG.pause = 0;              % pause between setups
CFG.preargs = {};           % any arguments to pass to the model BEFORE the integration config
CFG.postargs = {};          % any arguments to pass to the model AFTER the integration config
CFG.randseed = [];          % if random system/policy and repeatable results are needed

% process inputs
if nargin < 2, cfg = struct; end;
if nargin < 3, setups = SETUPS; end;
% parse config
cfg = parseconfig(cfg, CFG);
if ~iscell(cfg.preargs), cfg.preargs = {cfg.preargs}; end;
if ~iscell(cfg.postargs), cfg.postargs = {cfg.postargs}; end;
cfg.control = cfg.control && ~isempty(cfg.policy);

% evolution & runtimes storage
evol = struct;
times = struct;

% assign subplots to states
% create a dummy model for # of states and controls
m = feval(problem, 'model');
p = m.p; 
q = m.q * cfg.control;  % zero if no control
pq = p+q;
if pq <= 5;     % align on a single column
    h = pq; w = 1;
    figsize = [1000 800];
    subfigs = pq * 100 + 10 + (1:pq);
elseif pq <= 10;           % align on several columns
    figsize = [1100 970];
    w = ceil(pq / 5); h = ceil(pq / w);
    subfigs = h * 100 + w * 10 + (1:pq);
else
    figsize = [1100 970];
    w = ceil(pq / 5); h = ceil(pq / w);
    subfigs = (1:pq);
end;

% get initial state
if isempty(cfg.x0),
    try
        cfg.x0 = feval(problem, 'x0');
    catch
        cfg.x0 = zeros(p, 1);
    end;
end;

% create & setup figure
figh = figurex(figsize);
styles = {{'k--','LineWidth',2,'Color',[.6,.6,.6]}, {'k-','LineWidth',1}, ...
    {'k:','LineWidth',1,'Color',[0,1,0]}, {'k--','LineWidth',1,'Color',[0,0,1]}, ...
    {'k:','LineWidth',1,'Color',[1,0,0]}, {'k--','LineWidth',1,'Color',[.5,.5,0]}};

if isempty(cfg.randseed),
    seed = round(cputime*1000);
else seed = cfg.randseed;
end;

names = fieldnames(setups); Nsmax = -inf; leg = {};
for i = 1:length(names),
    sarg = setups.(names{i});
    if ~iscell(sarg), sarg = {sarg}; end;    % setup only contains one argument
    
    % create model
    m = feval(problem, 'model', cfg.preargs{:}, sarg{:}, cfg.postargs{:});
    
    % init history
    maxsteps = cfg.tend / m.Ts;
    t = (0 : maxsteps) * m.Ts;
    Ns = length(t) - 1;
    x = zeros(m.p, length(t)); x(:, 1) = cfg.x0;
    u = zeros(m.q, length(t)); % u(:, 1:end-1) = useq; 
    u(:, end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    rand('twister', seed);    % in case things depend on random events
    
    % samples loop
    tmark = cputime;
    for k = 1:Ns,
        % take action from policy if any specified
        if cfg.control,
            u(:, k) = feval(cfg.policy, x(:, k), (k-1) * m.Ts, cfg.policyargs{:});
        end;
        % apply to system
        [x(:, k+1) r(k+1) terminal] = feval(m.fun, m, x(:, k), u(:, k));
        if terminal, Ns = k; break; end;      % entered terminal state
    end;
    trun = cputime - tmark;     % runtime for simulation

    % form and save history
    hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1);    
    evol.(names{i}) = hist;
    times.(names{i}) = trun;
    Nsmax = max(Ns, Nsmax);     % last step in any run

    si = mod(i, length(styles)) + 1;
    % plot states each in its own subplot
    for ip = 1:p,
        if pq<10,
            subplot(subfigs(ip)); 
        else
            subplot(w,h,ip); 
        end;
        if ip > length(cfg.plotfuns),
           plot(hist.t, hist.x(ip, :), styles{si}{:}); hold on;
        else
           feval(cfg.plotfuns{ip}, hist.t, hist.x(ip, :), styles{si}{:}); hold on;
        end;
    end;
    % plot any controls
    for iq = 1:q,
        if pq<10,
            subplot(subfigs(p+iq)); 
        else
            subplot(w,h,p+iq); 
        end;
        if p+iq > length(cfg.plotfuns),
            stairs(hist.t, hist.u(iq, :), styles{si}{:}); hold on;
        else
            feval(cfg.plotfuns{iq}, hist.t, hist.u(iq, :), styles{si}{:}); hold on;
        end;
    end;
    
    leg{end+1} = [names{i} ', t=' num2str(trun*1000) ' msec'];

    % if configured to do so, pause after each but the last
    if cfg.pause && (i < length(names)), drawnow; pause; end;
end;

% add y labels
for ip = 1:p,
    if pq<10,
        subplot(subfigs(ip)); 
    else
        subplot(w,h,ip); 
    end;
    
    if isfield(m, 'xnames'),
        ylabel([m.xnames{ip} '(kT_S)']);
    else
        ylabel(['x_{' num2str(ip) '}(kT_S)']);
    end;
end;
for iq = 1:q,
    if pq<10,
        subplot(subfigs(p+iq)); 
    else
        subplot(w,h,p+iq); 
    end;    
    if isfield(m, 'unames'),
        ylabel([m.unames{iq} '(kT_S)']);
    else
        ylabel(['u_{' num2str(iq) '}(kT_S)']);
    end;
end;
% add legend on top of each column
% and time axis on bottom of each column
% for i = 1:w,
%     subplot(h, w, i);
%     lh = legend(leg, 'Location', 'Best');
%     set(lh, 'FontSize', cfg.legfontsize);
% end;
for i = pq-w+1:pq,
    subplot(h, w, i);
    xlabel('kT_S'); % legend(leg, 'Location', 'Best');
end;

% end ODETEST() returning EVOL, TIMES, FIGH