function x = unwrapangle(x, xs, xeps)
% Takes wrapped angle in [-pi, pi) and unwraps so that it forms a continuous trajectory
% Parameters:
%   X       - (vector) the trajectory
%   XS      - an approximate steady-state of the trajectory
%   XEPS    - discontinuity threshold, default 0.5
% A discontinuity is detected if the difference between two consecutive points on the traj is
% larger than 2*pi - xeps; so xeps must be carefully chosen given the speed of the dynamics
% if xeps is too small, discontinuities may go unnoticed; 
% if xeps is too large, false discontinuities may be detected where the state varies quickly

if nargin < 3, xeps = 0.5; end;

% find discontinuities
di = find(abs(diff(x)) > (2*pi-xeps));

% remove discontinuities by appropriately adding or subtracting 2*pi
for i = 1:length(di),
    dx = -sign(x(di+1) - x(di)) * 2*pi;
    x(di+1:end) = x(di+1:end) + dx;
end;

% finally, shift using an appropriate multiple of 2*pi
% so that final state is as close to xs as possible
dx = -2*pi * round((x(end) - xs) / (2*pi));
x = x + dx;

% done