function [xplus, rplus, terminal, G] = mararm_mg(m, x, u, mode, iagent, G, xplusothers)
%  Discrete-time dynamics of the Markov game for decoupled, homogeneous two-link robot arms.
%  [XPLUS, RPLUS, TERMINAL, G] = MALINEAR_MG(M, X, U, MODE, IAGENT, G)
%  Parameters:
%   M           - the model specification
%   X           - current state, x(k), as a flat vector
%   U           - commands u(k) as a flat vector
%       Both X and U should contain states for all agents (in 'real' mode) and for the current
%       agents and those connected to it (in 'pred' mode), see below
%   MODE        - 'real' for "real" complete transition (default value)
%                 'pred' for prediction from the perpective of a single agent; state and action should be
%                 arranged so that the first is of this agent, and then G
%   IAGENT      - in mode 'pred', index of this agent
%   G           - in mode 'pred', connections of this agent, as an adjacency row vector
%   XPLUSOTHERS - in mode 'pred', the next states of the other agents (to avoid simulating); 
%               if only some agent states are available, the others should be NaN
%               Note since the system is very simple, using this option will likely not pay off
%               Flat vector containing only the states of the neighbors
%  Returns:
%   XPLUS       - agent states at next sample, x(k+1)
%   RPLUS       - ensuing rewards, r(k+1)
%   TERMINAL    - whether the state is terminal (i.e., end of episode)
%               Should always be 0 for a continuing task
%   G           - graph of all agents (in 'real' mode) or connections of this agent ('pred' mode); 

if mode(1) == 'r',  nagents = m.nagents; G = m.G;
else                nagents = 1 + sum(G);
end;
if nargin < 7,      xplusothers = []; end;

x = reshape(x, m.p1, nagents); u = reshape(u, m.q1, nagents);
xplusothers = reshape(xplusothers, m.p1, []); % may be empty

% compute transition
xplus = nan(size(x));
for i = 1:nagents,
    % next state gets picked up from predictions if:
    % we are in prediction mode, and not handling the current agent, and a prediction is available
    if mode(1) == 'p' && i > 1 && ~isempty(xplusothers) && ~any(isnan(xplusothers(:, i-1))),
        xplus(:, i) = xplusothers(:, i-1);
    else    % otherwise, we must simulate
        if mode(1) == 'p' && i > 1, warning('simulating a neighbor'); end;
        xplus(:, i) = rarm_discretestep(m, x(:, i), u(:, i));
    end;
end;

% rewards

% in real mode, rewards must be found for all agents, in prediction mode, just for the first
if m.rewtype(1) == 'q', % normalized quadratic consensus reward
    if mode(1) == 'p',  % 'prediction' mode: diff between this agent and all neighbors
        % penalty is average quadratic distance to all neighbors, weighted by Q
        penalty = 0;
        for j = 2:nagents,
            dij = abs(xplus(:, 1) - xplus(:, j));
            % compute distance along SO1/circle, for the angular dimensions
            if dij(1) > pi, dij(1) = 2*pi - dij(1); end;
            if dij(3) > pi, dij(3) = 2*pi - dij(3); end;
            penalty = penalty + dij' * m.Q * dij;
        end;
        rplus = 1 - penalty / (nagents-1) / m.nonnorm_maxr;
    else                % 'real' mode: differences between all agents and their neighbors
        rplus = zeros(1, nagents);
        for i = 1:nagents,
            % move over this agent with neutral reward, if it doesn't have any neighbors
            if ~any(G(i, :)), rplus(i) = 1; continue; end;
            penalty = 0;
            for j = find(G(i, :)),
                dij = abs(xplus(:, i) - xplus(:, j));
                % compute distance along SO1/circle, for the angular dimensions
                if dij(1) > pi, dij(1) = 2*pi - dij(1); end;
                if dij(3) > pi, dij(3) = 2*pi - dij(3); end;
                penalty = penalty + dij' * m.Q * dij;
            end;
            rplus(i) = 1 - penalty / nnz(G(i, :)) / m.nonnorm_maxr;
        end;
    end;
elseif m.rewtype(1) == 'n',
    if mode(1) == 'p',  rplus = nan(1, nagents);
    else                rplus = nan;
    end;
else 
    error('Unknown reward type [%s]', m.rewtype);
end

xplus = xplus(:);
terminal = 0;

end         % MAIN FUNCTION END =================

