function varargout = cerbfqi(cfg)
% Cross-entropy optimization of approximator structure for
% Q-value iteration with linear, normalized RBF approximation.
%   VARARGOUT = CERBFQI(CFG)
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see he str2cfg
% Outputs:
%   PHISTAR, - TODO
%

if nargin < 1, cfg = struct(); end;

% TODO implement keepfields

% -----------------------------------------------
% Process configuration structure

% default config
% function config
CFG.run = 0;                        % run learning
CFG.resume = 0;                     % resume learning
CFG.replay = 0;                     % replay learned policy
CFG.stats = 0;                      % plot statistics
CFG.problem = '';                   % what problem to solve
CFG.datafile = 'ceqidata';          % save data to file
% main algorithm config
CFG.type = 'crad';                  % 'full' for searching centers and radii, 'crad' for constrained radii,
    % 'rej' for rejecting radii samples that do not fulfil the T&VR conditions
CFG.cost = 'mc';                    % 'mc' for monte-carlo, 'br' for Bellman residual
% CFG.crad_method = 'tvr';            % method for constraining radii: 'tvr' = Tsitsiklis Van Roy
%                                     % (option currently unused)
CFG.gamma = 0.95;                   % discount factor
CFG.N = 10;                         % number of basis functions to use
CFG.X0 = [];                        % (required) distribution of interesting initial states
CFG.U = [];                         % (required) flat action space, pxM
    % (X0 and U can also be supplied by the problem but if not user has to supply them)
% CE config
CFG.ce_N = -1;                      % number of samples to draw at each iteration; -1 for auto, N = min(c * Nphi, Nmax)
CFG.ce_maxN = 1000;                 % maximum value for N (cap)
CFG.ce_c = 3;                       % N = min(c * Nphi, Nmax), when auto
CFG.ce_rho = 0.05;                  % best samples at each iteration to use in updates
   % when < 1, the (100*(1-rho))th percentile of best samples will be used
   % when > 1, intended as an exact elite number of samples
CFG.ce_alpha = 1;                   % smoothed update coefficient for mean of params distribution
CFG.ce_beta = -1;                    % similar, for variance
CFG.ce_maxiter = 100;               % max # iterations for CE optimization
CFG.ce_d = 5;                       % how many iterations should the keep roughly constant to consider convergence
CFG.ce_eps = .01;                   % threshold for cost increase
CFG.ce_divpenalty = -1e10;          % score of diverging solutions, for just evaluating M-C leave at default []
% Q-iteration config
CFG.qi_eps = .01;                   % threshold for QI convergence
CFG.qi_maxiter = 300;               % max number of iterations for Q-iteration
% M-C simulation config for score evaluation
CFG.mc_maxsteps = 0.01;             % max steps in trajectory to compute returns
    % (when < 1, it's an error level from which the number of steps will be computed)
% LP config for score evaluation
CFG.radlp_eps = 0.001;
% replay config
CFG.interph = 0;                    % interpolated (averaged) policy
CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 30;                      % end time for replay
% stats & figure output config
CFG.plottarget = 'screen';          % 'screen', '', 'latex', or 'beamer'. If 'screen' figures will remain open
CFG.savetheta = 0;                  % save param history in stats
CFG.savedir = '';
CFG.savefig = '';
CFG.showdiverging = 0;              % whether to show convergence plot that includes diverging samples
% display config
CFG.verb = 3;                       % verbosity: the higher, the more detailed the messages displayed
CFG.noplot = 0;                     % whether to suppress figure plots
CFG.silent = 0;                     % suppress all output
CFG.iterdisp = 1;                   % feedback after every iterdisp iterations
CFG.itersave = 1;                   % save after each itersave iterations
CFG.sampledisp = 25;                % feedback after sampledisp samples evaluated in each iteration

% Early defaults (initialized before calling problem defaults)
ECFG.ce_params = {};                % extra parameters for problem calling in 'ce' mode
ECFG.model_params = {};             % extra parameters for problem calling in 'model' mode

% If caller provided string, parse it into a structure
if ischar(cfg),
    cfg = str2cfg(cfg, [fieldnames(CFG); fieldnames(ECFG)]);
end;
% Install 'early' defaults: those that the problem might depend on
cfg = checkparams(cfg, ECFG);
% Try installing problem defaults whenever caller does not specify values
% Problem must be specified for this to work
try     
    if ~isempty(cfg.problem), cfg = checkparams(cfg, feval(cfg.problem, 'ce', cfg.ce_params{:})); end;
catch
% 	dispx('Error in problem(''ce'')', cfg.verb, 0);
end;
% Install function defaults for everything else
cfg = checkparams(cfg, CFG);
% Process configuration dependencies
if cfg.silent, cfg.verb = -Inf; cfg.noplot = 1; end;
cfg.grayscale = grayscalefromconfig(cfg);

cfg.noinit = (cfg.resume || cfg.replay || cfg.stats) && exist([cfg.datafile '.mat'], 'file');

% -----------------------------------------------
% Data loading
if cfg.noinit,        % load data file, making sure that cfg is not overwritten
    cfg1 = cfg;
    load(cfg.datafile);
    % necessary overwrites from cfg vector with which algorithm was run
    cfg1.N = cfg.N;
    cfg = cfg1;
    clear cfg1;
    dispx(['Current data loaded from [' cfg.datafile '].'], cfg.verb, 1);
end;

% -----------------------------------------------
% Setup model
if ~cfg.noinit,
    model = feval(cfg.problem, 'model', cfg.model_params{:});    
end;

% Config settings for auto and derived parameters (require model so placed after model init)
% Number of structural parameters: variable-wise center and radius for each basis function
cfg.M = size(cfg.U, 2);
switch cfg.type,
    case {'full','rej'},
        cfg.Nphi = 2 * cfg.N * model.p * 2; % p*N center coordinates + p*N radii; for each 2 distr parameters
    case 'crad',
        cfg.Nphi = 2 * cfg.N * model.p;     % only p*N centers are independent, radii constrained by centers
end;
if cfg.ce_N < 0,                % auto sample set size
    cfg.ce_N = min(cfg.ce_c * cfg.Nphi, cfg.ce_maxN);
end;
if cfg.mc_maxsteps < 1,         % max steps is relative return error, use it to compute maxsteps
    cfg.mc_maxsteps = ceil(log(cfg.mc_maxsteps * (1-cfg.gamma) / model.maxr) / log(cfg.gamma));
end;
if cfg.ce_beta < 0, cfg.ce_beta = cfg.ce_alpha; end;

% Echo config
dispx('Cross-Entropy Q-iteration called with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1);

% Check file presence
if cfg.run && exist([cfg.datafile '.mat'], 'file'),
    reply = input(['File [' cfg.datafile '] already exists (possible overwrite). Continue? Y/N [N]: '], 's');
    if isempty(reply) || reply == 'N', return; end;
end;       

% -----------------------------------------------
% CE optimization of Q-iteration

if cfg.run || cfg.resume,
       
    % -----------------------------------------------
    % Do Q-iteration
    dispx('Performing C-E adaptation of linear Q-value iteration...', cfg.verb, 0);
    
    % -----------------------------------------------
    % Initialize if not resuming
    if ~cfg.resume,
        N = cfg.N;      % for easy access
        
%         rand('state', 1000);      % uncomment for reproducible results
%         randn('state', 1000);

        % with this RBFs are concentrated in the middle rather than uniformly distributed, 
        % but it's unclear how to deterministically place RBFs centers uniformly,
        % and I want to provide the same initial conditions to the algorithms for a more
        % meaningful comparison
        g.cmean     = zeros(model.p, N);
        % ensure most of the RBFs stay within the input space
        g.cstd      = repmat(model.maxx/2, 1, N);
        if strcmp(cfg.type, 'full') || strcmp(cfg.type, 'rej'),
            % RBFs as wide on average so as to have intersections between them at about 0.5
            g.radmean   = repmat(model.maxx/(2*(N^(1/model.p)-1)), 1, N);
            % allow width to vary between 0 and double
            g.radstd    = g.radmean/2;
        end;
        
        G{1} = g;
        
        % structural, linear parameters, score storage
        PHI = {}; THETA = {};
        J = NaN + zeros(cfg.ce_N, cfg.ce_maxiter); Jhat = NaN + zeros(1, cfg.ce_maxiter);
        % placeholder for phi, theta matrices, j vector at current iteration
        phi = zeros(cfg.ce_N, 2*model.p, N);    % centers and radii stacked on top of each other
    	theta = zeros(cfg.ce_N, N, cfg.M);
        j = zeros(cfg.ce_N, 1);
        flag = NaN + zeros(cfg.ce_N, cfg.ce_maxiter);
        
        k = 1;
        trun = 0;
        
        % init elements for Tsitsiklis & Van Roy conditions
        switch cfg.type,
            case 'rej',
                % init the bound on the sum of RBF activations for the centers
                % the bound for j ~= i is 1-gamma/2, but here we add one such that
                % we don't need to subtract the activation for j = i at each sample
                actbound = 1 + (1-cfg.gamma) / 2;
            case 'crad',
                % minimize 1/rad^2, i.e. maximize radii
                lp.f = ones(model.p, 1);
                % keeping 1/rad^2 positive, i.e. radius making sense
                lp.lb = zeros(model.p, 1) + cfg.radlp_eps;
    %             % and keeping 1/rad^2 at reasonable values
    %             lp.ub = zeros(model.p, 1) + 3000;
                lp.ub = [];
                % and ensuring that convergence constraints on the activation of the RBFs are
                % satisfied
                lp.b = zeros(N-1, 1) - log((1-cfg.gamma) / 2 / (N-1)) + cfg.radlp_eps;
                % Old version, technically unclear if correct, without /2 term -- still it never gave
                % divergence while being used
    %             lp.b = zeros(N-1, 1) - log((1-cfg.gamma) / (N-1)) + cfg.radlp_eps;
                % don't display anything (many, many lp problems will be solved)
                lp.opt = optimset('Display', 'off');  
                % matrix to store the solution of the LPs
                lp.sol = zeros(model.p, N);
        end;
        
        if strcmp(cfg.cost, 'br'),
            dispx('Precomputing transition and rewards table for X0xU0...', cfg.verb, 1);
            % Init transition and rewards table for set of interesting states
            S.X0 = cfg.X0; S.U0 = cfg.U;
            S.N0 = size(S.X0, 2); S.M0 = size(S.U0, 2);
            S.beta = ones(S.N0, S.M0) ./ (S.N0 * S.M0);
            % transition and reward table
            S.F = zeros(model.p, S.N0, S.M0); S.R = zeros(S.N0, S.M0);
            for i = 1:S.N0,
                for j = 1:S.M0,
                    [S.F(:, i, j) S.R(i, j)] = feval(model.fun, model, S.X0(:, i), S.U0(:, j));
                end;
            end;
            % make F size p x total number of samples for ease of use during iterations
            S.F = reshape(S.F, model.p, S.N0*S.M0);
            dispx('done.', cfg.verb, 2);
        end;
        
    end;        % initialization IF

    % -----------------------------------------------
    % Main loop of C-E optimization

    tmark = cputime;
    conv = 0;
    % preinit c, rad
    c = zeros(model.p, N); rad = zeros(model.p, N);
    while ~conv && k <= cfg.ce_maxiter,
        
        crejects = 0; radrejects = 0;
        % draw ce_N samples from the current distribution G{k}, each sample is c and rad
        for i = 1 : cfg.ce_N,            
            % draw centers ensuring they fall within state space
            for ii = 1:N,
                % draw #ii RBF center, rejecting any samples that exceed the state space boundary
                iic = g.cmean(:, ii) + g.cstd(:, ii) .* randn(model.p, 1);
                while any(abs(iic) > model.maxx), 
                    crejects = crejects + 1;
                    iic = g.cmean(:, ii) + g.cstd(:, ii) .* randn(model.p, 1);
                end;
                c(:, ii) = iic;
            end;
            % draw radii according to specified method (cfg.type)
            switch cfg.type,
                case 'rej',         % with sample rejection
                        % THIS METHOD IS UNFEASIBLE. too many rejects, very, very little chance
                        % of getting a good sample
                        rad = g.radmean + g.radstd .* randn(model.p, N);
                        % refuse sample if the radius is negative, or it
                        % doesn't satisy the Tsitsiklis & Van Roy condition
                        while (any(any(rad <= 0)) || any(sum(rbfvector(c, N, N, c, rad, 0), 2) >= actbound)) ...
                                && radrejects < 100000,
                            radrejects = radrejects + 1;
                            rad = g.radmean + g.radstd .* randn(model.p, N);
                        end;
                case 'full',        % draw from distribution, ensuring positivity
                    for ii = 1:N,
                        % reject samples with negative or zero components
                        iirad = g.radmean(:, ii) + g.radstd(:, ii) .* randn(model.p, 1);
                        while any(iirad <= 0), 
                            radrejects = radrejects + 1;
                            iirad = g.radmean(:, ii) + g.radstd(:, ii) .* randn(model.p, 1);
                        end;
                        rad(:, ii) = iirad;
                    end;                    
                case 'crad',        % constrained radius by T & VR method
                    for ii = 1:N, 
                        % compute constraint matrix D of size (N-1) x p, 
                        % other RBFs on rows, state variable index on columns (that's why
                        % the diff matrix is transposed)
                        lp.D = (repmat(c(:, ii), 1, N-1) - c(:, [1:ii-1 ii+1:end]))' .^ 2;
                        % Variant 1: maximize radii
                        [lp.sol(:, ii) lplambda lpflag] = linprog(lp.f, -lp.D, -lp.b, [], [], lp.lb, lp.ub, [], lp.opt);
                        % Variant 2: maximize activations of this RBF by the other centers
%                         [lp.sol(:, ii) lplambda lpflag] = linprog(sum(lp.D), -lp.D, -lp.b, [], [], lp.lb, lp.ub, [], lp.opt);
                        if lpflag < 0, 
                            dispx(['LP did not converge! k=' num2str(k) ' samp#=' num2str(i) ' rbf#=' num2str(ii) ...
                                ' lpflag=' num2str(lpflag) ' 1/rad^2=' num2str(reshape(lp.sol(:, ii), 1, []))], cfg.verb, 3);
                        end;
                    end;
                    rad = 1 ./ sqrt(lp.sol);
            end;
            
            % run Q-iteration
            [th, flag(i, k)] = rbfqiter(c, rad, model, cfg);
            
            if flag(i, k) < 0 && ~isempty(cfg.ce_divpenalty),           % if divergence, penalize
                j(i) = cfg.ce_divpenalty;
            else
                % evaluate result
                switch cfg.cost,
                    case 'mc',      % Monte-Carlo evaluation of set of interesting states X0
                        j(i) = mc_rbfq(c, rad, squeeze(theta(i, :, :)), model, cfg.X0, cfg);
                    case 'br',      % Bellman residual on set of samples X0 x U
                        % lhs = Qhat(x0, u0)
                        % REMARK when U ~= U0, correct indices must be picked up from theta
                        lhs = rbfvector(S.X0, S.N0, N, c, rad) * th;
                        % rhs = rho(x0, u0) + gamma max_u' Qhat(f(x0, u0), u')
                        rhs = S.R + cfg.gamma * reshape(max(nrbfvector(S.F, S.N0*S.M0, N, c, rad) * th, [], 2), S.N0, S.M0);
                        % (code for maximization problem so score is multiplied by -1)
                        j(i) = -sum(sum(S.beta .* (lhs - rhs).^2));
                end;
            end;
            
            % store RBF parameters and linear weights
            phi(i, :, :) = [c; rad];
            theta(i, :, :) = th;
            % visual feedback of algorithm progress
            if ~mod(i, cfg.sampledisp),
                dispx([num2str(i) ' out of ' num2str(cfg.ce_N) ' samples evaluated'], cfg.verb, 3);
            end;
        end;
        
        % sort scores and parameter sets with best first
        [j ind] = sort(j, 1, 'descend');
        phi = phi(ind, :, :);
        theta = theta(ind, :, :);
        flag(:, k) = flag(ind, k);
        % compute last elite sample
        if cfg.ce_rho < 1,      % use percentile
            Jhat(k) = prctile(j, 100*(1-cfg.ce_rho));
            Ne = find(j >= Jhat(k), 1, 'last');
        else                    % use fixed number of elite samples
            Ne = cfg.ce_rho;
            Jhat(k) = j(Ne);	% worst performance in elite samples
        end;
        % compute update of distribution as sampled mean and variance of elite samples
        % (general form of smoothed update)
        % note standard deviations are computed by dividing by number of samples (second input to std flag=1)
        % as opposed to (number of samples - 1)
        g.cmean = (1-cfg.ce_alpha)   .* g.cmean      + cfg.ce_alpha  .* squeeze(mean(phi(1:Ne, 1:model.p, :)));
        g.cstd = (1-cfg.ce_beta)     .* g.cstd       + cfg.ce_beta   .* squeeze(std(phi(1:Ne, 1:model.p, :), 1));
        if strcmp(cfg.type, 'full'),
            g.radmean = (1-cfg.ce_alpha) .* g.radmean    + cfg.ce_alpha  .* squeeze(mean(phi(1:Ne, model.p+1:end, :)));
            g.radstd = (1-cfg.ce_beta)   .* g.radstd     + cfg.ce_beta   .* squeeze(std(phi(1:Ne, model.p+1:end, :), 1));
        end;

        G{k+1} = g;
        
        % save data from current iteration
        J(:, k) = j; 
        PHI{k} = phi;
        THETA{k} = theta;
        % update stats
        trun = trun + (cputime - tmark);
        
        % check convergence condition
        deltaj = diff( Jhat(max(1, k - cfg.ce_d):k) );
        conv = k > cfg.ce_d && all(0 <= deltaj) && all(deltaj <= cfg.ce_eps);
        
        % visual feedback of algorithm progress
        if ~mod(k, cfg.iterdisp),
            dispx(['k=' num2str(k) ' C-E iter done, d(Jhat(k-d:k))=[' num2str(deltaj) ...
                '], J*=' num2str(max(max(J))) ', %div=' num2str(sum(flag(:, k) < 0)/cfg.ce_N * 100) ...
                ', #rej=' num2str(crejects) '+' num2str(radrejects)], cfg.verb, 2);
        end;
        % data backup
        if ~mod(k, cfg.itersave),
            save(cfg.datafile);
            dispx(['Data at iter k=' num2str(k) ' saved to [' cfg.datafile '].'], cfg.verb, 1);
        end;
        
        tmark = cputime;
        k = k + 1;
    end;        % WHILE not converged and allowed more iterations

    if conv,	dispx('Convergence detected. Algorithm stopped.', cfg.verb, 0);
    else        dispx(['maxiter=' num2str(cfg.ce_maxiter) ' exhausted. Algorithm stopped'], cfg.verb, 0);
    end;
    
    % return best structural and linear parameter vectors encountered so far
    Jstar = max(max(J));
    [istar, kstar] = find(J == Jstar, 1);
    phistar = squeeze(PHI{kstar}(istar, :, :));
    thetastar = squeeze(THETA{kstar}(istar, :, :));
end;


% -----------------------------------------------
% Replay
if cfg.replay,

    % structure params
    cstar = phistar(1:model.p, :);
    radstar = phistar(model.p+1:end, :);
    
    % compute locally optimal policy (local optimal discrete action for each basis
    % function)
    if cfg.interph
        [Qstar ui] = max(thetastar, [], 2); clear Qstar;
        hstar = cfg.U(:, ui)';
    end;

    % initial state
    if ~isempty(cfg.x0), x0 = cfg.x0(:);      % specified initial state
    else x0 = zeros(model.p, 1); end;         % zeros
    dispx(['Controlling from x0=' num2str(reshape(x0, 1, [])) ], cfg.verb, 0);

    % init history
    t = 0 : model.Ts : cfg.tend;
    Ns = length(t)-1;       % number of samples / time instances at which control is applied
    x = zeros(model.p, length(t)); x(:, 1) = x0;
    u = zeros(model.q, length(t)); u(:, end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    % steps loop
    for k = 1:Ns,
        % compute optimal action
        if cfg.interph,     % either interpolated
%             size(nrbf(x(:, k), cfg.N, cstar, radstar))
%             size(hstar)
            u(:, k) = nrbf(x(:, k), cfg.N, cstar, radstar) * hstar;
        else                % or crisp (ties broken by choosing first best action)
            Qa = squeeze(nrbf(x(:, k), cfg.N, cstar, radstar)) * thetastar;     % TODO ?? why squeeze needed?
            u(:, k) = cfg.U(:, find(Qa == max(Qa), 1));
        end;
        % apply to system
        [x(:, k+1) r(k+1) terminal] = feval(model.fun, model, x(:, k), u(:, k));
        if terminal, Ns = k; break; end;      % entered terminal state
    end;
 
    % plot history & optionally save figures
    hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1);
    if ~cfg.noplot,
        toscreen = strcmp(cfg.plottarget, 'screen') || isempty(cfg.plottarget);
        figh = plothistory(hist);
        saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
        if isfield(model, 'plotfun'),
            mpfigh = feval(model.plotfun, hist);
            if ~isempty(mpfigh), figh(end+1) = mpfigh; end;
        end;
        if ~toscreen, close(figh); end;
    end;

end;        % IF replay

% -----------------------------------------------
% Plot statistics
if cfg.stats && ~cfg.noplot, 
    % compute mean and std of cost excluding diverging samples
    K = k - 1;
    mJ = zeros(1, K); stdJ = mJ; dimJ = mJ;
    for ik = 1:K,
        if exist('flag'),
            % WARNING some old data files have unsorted flags, and -10 divergence score
            nondiv = find(flag(:, ik) >= 0);
        else 
            % use this second form when working with data files where bug of not sorting the flags is
            % fixed
            nondiv = find(J(:, ik) > cfg.ce_divpenalty);
        end;
        mJ(ik) = mean(J(nondiv, ik));
        stdJ(ik) = std(J(nondiv, ik), 1);
        dimJ(ik) = length(nondiv);
    end;

    % grayscale styles
    gs.mean = {'k'};
    gs.max = {'k', 'LineWidth', 2};
    gs.quantile = { 'k--', 'LineWidth', 2};
    % black to dirty white
    gs.cm = gray(96); gs.cm = gs.cm(24:end);
    cs.mean = {'k'};
    cs.max = {'k', 'LineWidth', 2};
    cs.quantile = { 'k--', 'LineWidth', 2};
    cs.cm = jet(64);
    % set style
    if cfg.grayscale, sty = gs; 
    else sty = cs; end;    
    
    % readable labels
    commonprop = {'Interpreter', 'Latex', 'FontSize',13};
    rl.avg = '$\frac{\sum_l s(\xi_l)}{N_\xi}$'; 
    rl.quantile = '$\lambda_t$'; 
    rl.max = '$\max_l s(\xi_l)$';
    rl.x = {'Iteration t'};
    rl.y = {'Score s'};
    labels = rl;    % no psfrag labels

    toscreen = strcmp(cfg.plottarget, 'screen') || isempty(cfg.plottarget);
    figh = figure; hold on; 

    errorbar(1:K, mJ, stdJ, sty.mean{:});  
    plot(1:K,max(J(:, 1:K)), sty.max{:});
    plot(1:K, Jhat(1:K), sty.quantile{:});
    xlabel(labels.x{:}); ylabel(labels.y{:}); 
    lh = legend(labels.avg, labels.max, labels.quantile); 
    set(lh, 'Location', 'Best', commonprop{:});
    if toscreen, title(['Non-diverging samples, N=' num2str(cfg.N)]); end;
    set(figh, 'Name', cfg.datafile, 'NumberTitle', 'off');
    
    saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
    if ~toscreen, close(figh); end;
    dimJ
end;

% -----------------------------------------------
% Backup data
if cfg.run || cfg.resume, 
    % save into the same directory as the problem
    savedir = fileparts(which(cfg.problem));
    % if default file name is used, generate a unique file name
    if strcmp(cfg.datafile, CFG.datafile),
        c = fix(clock);
        for i = 1:length(c),
            cfg.datafile = [cfg.datafile '-' num2str(c(i), '%02d')];
        end;
        delete([CFG.datafile '.mat']);
    elseif ~strcmp(savedir, pwd)
        delete([cfg.datafile '.mat']);   % need to save in a different directory, so delete anyway
    end;    % otherwise just overwrite
    cfg.datafile = [savedir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Cross-entropy Q-iteration finished. Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;


% -----------------------------------------------
% set output
if cfg.run || cfg.resume,       % output optimal parameter vectors, optimal score
    varargout = {phistar, thetastar, Jstar};
elseif cfg.replay               % output history and possibly figure handles
    fig = ~cfg.noplot && (strcmp(cfg.plottarget, 'screen') || isempty(cfg.plottarget));
    if fig,         varargout = {hist, figh};
    else            varargout = {hist, []};
    end;
end;


% END fuzzyqi() RETURNING varargout =================================================================


