function lspihfindosc(ocfg)
% Find oscillating experiments for LSPIH. Processes experiments that did not check for
% oscillations (should not be needed any longer)

if nargin < 1, ocfg = struct(); end;

% ==== DECLARE CONFIGURATION DEFAULTS ====
OCFG.resave = 0;                    % save over old datafile
OCFG.datafile = [];                 % data file
OCFG.maxper = 10;                   % maximum period to check for (# iter)
OCFG.nper = 1;                      % number of periods to check for
OCFG.epsper = 1e-2;                 % maximum abs diff between two par vectors in diff periods
OCFG.checkq = 1;                    % also verify if the Q parameter is oscillating

ocfg = parseconfig(ocfg, OCFG);

if ~exist([ocfg.datafile '.mat'], 'file'),
    error(['File [' ocfg.datafile '.mat] does not exist. Terminating.']);
end;

load(ocfg.datafile);
if ~exist('htheta', 'var'),
    disp(['LSPIHFINDOSC: The algorithm used for [ ' ocfg.datafile ']' ...
        13 ' was not LSPIH. Will not change file.']);
    return;
end;

K = find(~isnan(deltah), 1, 'last');

% hist(1) == 0, so we start checking params from k = 2
for ith = 1 : K,  % the first index for which a 0-period (constant theta) is meaningful
    
    per = 2; osc = 0;   % don't test for period = 1, that's taken care of by the convergence test
    % try periods up to maxper, 
    % as long as oscillation wasn't detected (then it makes no sense to try longer periods)
    % and provided that the number of iterations allows for it
    while ~osc && (per <= ocfg.maxper) && (per * (ocfg.nper+1) < ith)
        
        osc = 1;                % start by assuming oscillating; tests below might falsify that
        for i = 0:per-1,                % i loops through the offsets of the iters that have to be equal
            for j = 1:ocfg.nper,                % make sure oscillation happens for at least nper periods
                osc = osc & all(abs(hthetah{ith-i} - hthetah{ith-j*per-i}) <= ocfg.epsper);
                % also verify value fun parameters if required
                if ocfg.checkq,
                    osc = osc & all(abs(thetah{ith-i} - thetah{ith-j*per-i}) <= ocfg.epsper);
                end;
%                 osc = osc & (norm(hthetah{ith-i} - hthetah{ith-j*per-i}) <= ocfg.epsper);
                if ~osc, break; end;            % don't test farther periods
            end;
            if ~osc, break; end;        % don't test other offets
        end;
        
        
        per = per + 1;
    end;
    per = per - 1;  % to revert the last increment
    
    if osc,
        kosc = ith - 1;
        disp(['[' ocfg.datafile ']: ' 13 '   Solution oscillates with period ' num2str(per)...
            '. Detected at iteration ' num2str(kosc)]);
        break;
    end;
end;

if osc && ocfg.resave,
    if exist('lspihfindosc_run', 'var') && lspihfindosc_run,
        disp(['LSPIHFINDOSC was already run for [' ocfg.datafile ']. Will not change file.']);    
    else
        disp(['Updating data and resaving [' ocfg.datafile ']']);
        % find iteration that is currently last
        K = find(~isnan(deltah), 1, 'last') - 1;
        % copy new configuration fields
        cfg = copyfields(ocfg, cfg);
        % remove the histories from kosc onwards
        for k = kosc+1:K,
            thetah{k+1} = []; hthetah{k+1} = [];
            deltah(k+1) = NaN; hdeltah(k+1) = NaN;
            condA(k) = NaN;
            if cfg.storeab, Ah{k} = []; bh{k} = []; end;
        end;
        % recompute the runtime approximately considering the removed iterations
        trun = trun * kosc / K; trunapprox = 1;  % flag
        % set current params
        theta = thetah{kosc+1}; htheta = hthetah{kosc+1};
        % set iterations counter
        k = kosc + 1;
        % set convergence flag to mark "oscillations"
        conv = -1;
        % just to re-emphasize: we are oscillating
        osc = 1;
        % flag ran
        lsiphfindosc_run = 1;

        % finally, rename old file as backup
        [status, output] = system(['rename ' ocfg.datafile '.mat ' ocfg.datafile '_backup.mat']);
        if status ~= 0, disp(output); error('Rename failed.'); end;
        % and save the updated data
        save(ocfg.datafile);
    end;
end;


