function  varargout = opddwelltime(xroot, model, cfg)
% OPD Optimistic planning for deterministic systems, returning full sequences
% Inputs:
%   xroot       - root state of current agent, the initial step is denoted step 0
%   Xneighb     - states of neighboring agents, as a cell array of matrices, the jth matrix contains
%       the predicted state trajectory of the jth neighbor, one state per column. The first column
%       should correspond to step 0. The lengths of the trajectories may be different for each
%       neighbor
%   model, cfg  - self explanatory
% Outputs:
%   u           - first action on greedy path
%   useq        - sequence of actions corresponding to greedy node, starting with u_0 and ending
%       at u_d-1 with d the depth of the greedy node
%   xseq        - corresponding sequence of states, x_0 to x_d
%   planinfo    - a structure with various information about the planning process, see code
%   plandata    - the actual planning tree developed

% Author: Lucian Busoniu

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % specify either expansion budget
CFG.ntrans = [];                % or number of simulated transitions
CFG.depth = [];                 % or target depth
CFG.criterion = 'optim';        % expansion criterion: optim(istic) i.e. regular OPD or unif(orm) expansion
CFG.U = [];                     % discrete, flat action space
CFG.uprev = [];                 % previous action (to ensure the dwell time in closed loop)
CFG.k = -1;                     % discrete time step at which the algo is called
CFG.dwelltime = 1;
CFG.actiontype = 'one';         % 'one' or 'seq' (return one action or full sequence)
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma; criterion = cfg.criterion; 
U = cfg.U;
M = size(U, 2);      % number of discrete actions
dimx = model.p;      % number of state vars

% start counting time
t = cputime;

if ~isempty(cfg.n), maxtreesize = M*cfg.n + 1; else maxtreesize = M*1000 + 1; end;
% prepare tree data 
Tparent = zeros(2, maxtreesize, 'uint32');        % parent and parent action
Tchildren = zeros(M, maxtreesize, 'uint32');      % child index for each action
Tdepth = zeros(1, maxtreesize, 'uint16');         % depth (uint16 as depth > 65536 considered unfeasible)
Tleaf = false(1, maxtreesize);                    % leaf flag
Tx = NaN(dimx, maxtreesize, 'double');            % state
Tr = NaN(1, maxtreesize, 'double');               % reward
TR = NaN(1, maxtreesize, 'double');               % partial return up to node (cumulative reward)
Tdwell = zeros(1, maxtreesize, 'uint8');          % dwell times (# steps since last switch, 255 is enough)

% initialize the root node (note 0 = null pointer)
root = 1;
Tparent(:, root) = 0;
Tchildren(:, root) = 0;
Tdepth(root) = 0;
Tleaf(root) = true;
Tx(:, root) = xroot;
Tr(root) = 0;
TR(root) = 0;
Tdwell(root) = cfg.dwelltime;           % root assumed to satisfy the constraint
treesize = 1;
B = 1 ./ (1-discount);

% 2) Main tree construction loop
tau = 0; stopflag = false;
while ~stopflag,
    
    % 2a) select candidate leaves for expansion
    leaves = find(Tleaf);
    if criterion(1) == 'o', 
        % optimistic: max b-value
        leaves = leaves(B == max(B));
    elseif criterion(1) == 'u',
        % uniform: smallest depth
        leaves = leaves(Tdepth(leaves) == min(Tdepth(leaves)));
    else
        error('Invalid expansion criterion');
    end;
    % break any ties between remaining candidates deterministically
    nodetoexpand = leaves(1);
    
    % 2b) expand the chosen node
    % if the current dwell time is too small, parent action must be kept constant
    if Tdwell(nodetoexpand) < cfg.dwelltime,    candidateu = Tparent(2, nodetoexpand);
    else                                        candidateu = 1:M;
    end;
    Tleaf(nodetoexpand) = false;        % no longer a leaf
    % run through the allowed actions
    for j = candidateu,
        % find index where new node will be, save it on the children table of current node
        child = treesize + 1;
        Tchildren(j, nodetoexpand) = child;
        
        % simulate transition and add data to tree
        [Tx(:, child), Tr(child)] = feval(model.fun, model, Tx(:, nodetoexpand), U(:, j));
        if Tr(child) < 0 || Tr(child) > 1, error('Rewards must be in [0, 1]'); end;
        
        % create new node
        Tparent(:, child) = [nodetoexpand; j];      % parent is expanded node, along jth action
        Tchildren(:, child) = 0;                    % init children pointers to "null"
        Tdepth(child) = Tdepth(nodetoexpand) + 1;   % one deeper than the parent
        Tleaf(child) = true;                        % new child is a leaf
        TR(child) = TR(nodetoexpand) + discount^double(Tdepth(nodetoexpand)) * Tr(child);
        % if action same as parent (or we are at start of traj), dwell time increases, else it's reset to 1
        % note the root is special case
        if nodetoexpand == root,
            if cfg.k == 1,
                if cfg.uprev == 0,      Tdwell(child) = 1;                      % any first action is a switch
                elseif cfg.uprev == -1, Tdwell(child) = Tdwell(root) + 1;       % first action not a switch
                end;
            else
                % NOTE the change to uprev = containing real action was not verified (included 3
                % sept 2015)
                if j == findflat(cfg.uprev, cfg.U),     Tdwell(child) = Tdwell(nodetoexpand) + 1;
                else                                    Tdwell(child) = 1;
                end;
            end;
        else
            if j == Tparent(2, nodetoexpand),   Tdwell(child) = Tdwell(nodetoexpand) + 1;
            else                                Tdwell(child) = 1;
            end;
        end;
            
        % we just added a new node, increase tree size
        treesize = treesize + 1;
    end;
    
    if cfg.k == 287 && tau <= 3,
    end;
    
    % 2c) update b-values (globally to keep the implem simple)
    B = TR(leaves) + discount.^double(Tdepth(leaves)) ./ (1-discount);

    % increase number of expanded nodes
    tau = tau + 1;
    % stop when budget (n or ntrans) exhausted or a node at target depth was expanded
    stopflag = (~isempty(cfg.n) && tau >= cfg.n) ...
        || (~isempty(cfg.ntrans) && (treesize - 1) >= cfg.ntrans) ...
        || (~isempty(cfg.depth) && Tdepth(nodetoexpand) == cfg.depth);
                
end;    % end main tree construction loop


% choose a leaf with maximal u-value (i.e., truncated return)
leaves = find(Tleaf); [Rmax, jmax] = max(TR(leaves)); node = leaves(jmax); 
% navigate greedy path upwards in the tree until level 1, just after root, possibly remembering the sequence
if cfg.actiontype(1) == 'o',    % 'one', root action      
    while Tparent(1, node) ~= root, node = Tparent(1, node); end;
    % choose the action that was taken at root to arrive at this level-1 node
    ustar = U(:, Tparent(2, node));
else                            % 'seq', entire sequence of actions
    % 4) Find action at the tree root
%     maxpath = node;
    while node ~= root,
        xseq(1:dimx, Tdepth(node)+1) = Tx(:, node);
        ustar(1:model.q, Tdepth(node)) = U(:, Tparent(2, node));
        node = Tparent(1, node);
%         maxpath = [node maxpath];
    end;
    xseq(1:dimx, 1) = Tx(:, root); % the first state is always the root one
end;

% output stats if requested
if nargout >= 2,
    planinfo.n = tau;
    planinfo.bounds = [Rmax max(B)];
    planinfo.depth = max(Tdepth);
    planinfo.trun = cputime - t;
    planinfo.treesize = treesize;
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        tau, planinfo.depth, planinfo.treesize); 
end;
% output tree data 
if nargout >= 3,
    plandata.Tparent = Tparent(:, 1:treesize);
    plandata.Tchildren = Tchildren(:, 1:treesize);
    plandata.Tdepth = Tdepth(1:treesize);
    plandata.Tleaf = Tleaf(1:treesize);
    plandata.Tx = Tx(:, 1:treesize);
    plandata.Tr = Tr(1:treesize);
    plandata.TR = Tr(1:treesize);
    plandata.Tdwell = Tdwell(1:treesize);
end;

% construct the output based on calling mode
switch nargout,
    case 1, varargout = {ustar};
    case 2, varargout = {ustar, planinfo};
    case 3, varargout = {ustar, planinfo, plandata};
    case 4, 
        if cfg.verb >= 3, fprintf('ustar=%s\n', num2strx(ustar, 'vector')); end;
        varargout = {ustar, xseq, planinfo, plandata};
end;


end % main function
