function varargout = llrfittedqi(cfg)
% Fitted Qs-iteration with local linear regression.
%   APPROX = FITTEDQI(CFG)              - in 'run' mode
%   [HIST, FIGH] = FITTEDQI(CFG)        - in 'replay' mode
%   FIGH = FITTEDQI(CFG)                - in 'sol' or 'evol' modes
% An implementation of fitted Q-iteration with local linear regression
% approximation.
%
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see str2cfg
% 
% Outputs:
%       APPROX      - final Q-function & policy approximator
%                   (non-parametric)
%       HIST        - the replay history (trajectories)
%       FIGH        - handles to the figures created

% TODO proper handling of terminal states

% Author: Lucian Busoniu
% Version: 
% History:

if nargin < 1, cfg = struct(); end;

% ==== DECLARE CONFIGURATION DEFAULTS ====
% function config
CFG.run = 0;                        % run learning
CFG.resume = 0;                     % resume learning
CFG.replay = 0;                     % replay learned policy 
CFG.sol = 0;                        % plot solution
CFG.evol = 0;                       % plot evolution (once every evol iterations)
CFG.datafile = 'llrfittedqidata';   % save data to file
CFG.datadir = [];                   % data dir
% main algorithm config
CFG.problem = '';                   % what problem to solve
CFG.gamma = [];                     % discount factor
CFG.U = [];                         % discrete action space (cell array of grids, or flat)
% LLR config
CFG.knn = 10;                       % # of samples
CFG.S = 'auto';                     % distance scaling: [] = no scaling, dim(x)-vector = diagonal scaling
                                    % 'auto' choose to scale state variables equally
CFG.wfun = 'unif';                  % weighting fun (kernel): [tri]'cube', 'quad', 'unif', 'triang'
CFG.knnmethod = 'jit';              % method for knn search, 'kdtree' (CPP) or 'jit' (Matlab JIT-optimized)
CFG.linsolvemethod = 'pinv';        % only 'pinv' supported (cannot use mldivide with precomputation)
CFG.extrap = 'sat';                 % extrapolation handling: 'sat' = saturate, 'none' = leave alone
% samples, iterations etc.
CFG.samples = [];                   % collection of samples object (or config for generate_samples)
CFG.loadsamples = [];               % load samples from this file
CFG.maxiter = 500;                  % max # of iterations to run
CFG.eps = 0.1;                      % convergence threshold
CFG.maxdelta = 1e15;                % divergence threshold
CFG.checkosc = 1;                   % whether to check for oscillations
CFG.maxper = 10;                    % osc: maximum period to check (# iter) -- >= 2
CFG.nper = 1;                       % osc: number of periods to check
CFG.epsper = 0.1;                   % osc: maximum abs diff between two par vectors in diff periods
% replay config
CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 30;                      % end time for replay
% stats config
CFG.storehistory = 1;               % whether to store solution history at every iteration
% display config
CFG.visualize = 0;                  % visualization level (0 = none, 1 = iteration-level)
CFG.viscfg = struct;                % visualization config options
CFG.itervis = 10;                   % visualization period
CFG.verb = 3;                       % verbosity: the higher, the more detailed the messages displayed
CFG.iterdisp = 10;                   % feedback after every iterdisp iterations
CFG.itersave = 50;                  % save after each itersave iterations
CFG.sampledisp = .05;               % feedback once this % of samples processed
CFG.silent = 0;                     % suppress all output
% figure config
CFG = setfigprop(CFG, 'addfields');
CFG.slice = [];                     % slice of solution to plot
CFG.miscinfo = '';                  % field for miscellaneous information

% Early defaults (initialized before calling problem defaults)
ECFG.model_params = {};             % parameters for problem calling in 'model' mode
ECFG.fittedqi_params = {};          % parameters for problem calling in 'fittedqi' mode

% List of fields that cannot change in-between iterations
KEEPFIELDS = {'problem', 'gamma', 'samples'};

% global variable that stores KDTREE_POINTERS pointers, for deletion purposes
global KDTREE_POINTERS;

% ==== PARSE AND PROCESS CONFIG ====
cfg = parseconfig(cfg, CFG, ECFG, 'fittedqi');
% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% If running, check presence of data file, warn on possible overwrite
if cfg.run && exist([cfg.datafile '.mat'], 'file'),
    reply = input(['File [' cfg.datafile '.mat] already exists (possible overwrite). Continue? Y/N [N]: '], 's');
    if isempty(reply) || reply == 'N', return; end;
end;       
% determine if initialization needed; if not, need to load: check if datafile exists
cfg.init = ~(cfg.resume || cfg.replay || cfg.sol || cfg.evol);
if ~cfg.init && ~exist([cfg.datafile '.mat'], 'file'),
    error(['File [' cfg.datafile '.mat] does not exist. Terminating.']);
end;

% ==== IF NOT INITIALIZING: NEED TO LOAD DATA ====
if ~cfg.init,        % load data file, making sure that cfg and KEEPFIELDS is not overwritten
    % optimize the loading time: only load large variables when needed
    dfv = who('-file', cfg.datafile);
    % REMARK remove here any large variables unnecessary for the current mode 
    if ~cfg.evol, dfv = rmstring(dfv, 'Qsh'); end;
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile, dfv{:});
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['LLRFittedQI: data loaded from [' cfg.datafile '].'], cfg.verb, 1);
%     approx = revise_approx(approx);             % for replay, etc functions that assume approx is initialized
    cfg.samples = revise_samples(cfg.samples);  % the same for samples
    % Rebuild trees
    if strcmp(cfg.knnmethod, 'kdtree'),
        dispx('Rebuilding kd-trees...', cfg.verb, 1);
        kdtrees = nan(M, 1);
        for j = 1:M, kdtrees(j) = kdtree_build(Xs(:, Is{j})'); end;
        % also add them to the approximator
        approx.kdtrees = kdtrees;
        KDTREE_POINTERS(end+1:end+M) = kdtrees;
        dispx([8 ' done.'], cfg.verb, 1);
    end;
end;

% Some error checking
if ~strcmp(cfg.linsolvemethod, 'pinv'), 
    error('Linear solving method [%s] unknown (only [pinv] unsupported).', cfg.linsolvemethod); 
end;
switch cfg.knnmethod,
    case 'kdtree',
        if ~(isempty(cfg.S) || all(cfg.S == 1)), 
            error('Distance scaling cannot be used for [kdtree].'); 
        end;
    case 'jit', 
    otherwise, error('kNN method [%s] unknown.', cfg.knnmethod); 
end;
if ~any(strcmp(cfg.extrap, {'sat', 'none'})), 
    error('Extrapolation handling method [%s] unknown.', cfg.extrap); 
end;

% Echo config
dispx('LLRFittedQI will run with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');

% ==== IF INITIALIZING: CREATE MODEL, AND IF NEEDED SAMPLES ====
if cfg.init,
    model = feval(cfg.problem, 'model', cfg.model_params{:});
    % samples object
    if ~isempty(cfg.loadsamples) && exist([cfg.loadsamples '.mat'], 'file'),
        % try loading samples -- performs no checking whether the same type
        % & number of samples is required
        loadwithprefix(cfg.loadsamples, 's_', 'cfg');
        cfg.samples = s_cfg.samples;
        clear s_cfg;
        dispx(['Samples object loaded from [' cfg.loadsamples ']'], cfg.verb, 1);
    else
        % heuristic check if we already have a samples collection supplied
        if isstruct(cfg.samples) && isfield(cfg.samples, 'N') && isfield(cfg.samples, 'X') ....
                && isfield(cfg.samples, 'U') && isfield(cfg.samples, 'Xp') ...
                && isfield(cfg.samples, 'R'),
                dispx('Samples object supplied', cfg.verb, 3);
                % nothing to do
        else    % generate samples using cfg.samples as the config
            dispx('Generating samples...', cfg.verb, 1);
            % use a placeholder for the approx in order to pass the
            % discrete actions to the generator function
            appdummy = struct; appdummy.U = cfg.U;
            cfg.samples = generate_samples(model, cfg.samples, appdummy);
            dispx([8 ' done.'], cfg.verb, 1);
        end;
    end;
    if isempty(cfg.U), error('LLRFittedQI requires a predefined, discrete action space.'); end;
end;

% ==== RUN FITTED QI ====
if cfg.run || cfg.resume,
       
    % Initialize if not resuming
    if cfg.init,
        % shorthand variables
        Ns = cfg.samples.N; 
        Xs = cfg.samples.X; Us = cfg.samples.U; 
        Xps = cfg.samples.Xp; Rs = cfg.samples.R; 
        p = model.p;
        knn = cfg.knn;
        switch cfg.extrap,
            case 'none',    satflag = 0;
            case 'sat',     satflag = 1;
        end;

        % retrieve action space
        if iscell(cfg.U),   U = flat(cfg.U); 
        else                U = cfg.U;
        end; 
        M = size(U, 2);                         % number of discrete actions

        % history
        deltah = nan(cfg.maxiter, 1); 
        if cfg.storehistory || cfg.checkosc, Qsh = cell(cfg.maxiter, 1); end;
        
        % execution time statistics
        timestat = struct;
        timestat.init = 0;

        % initialize distance scaling
        if isempty(cfg.S),
            cfg.S = ones(p, 1);
        elseif ischar(cfg.S) && strcmp(cfg.S, 'auto'),
            if isfield(model, 'minx'),      cfg.S = 1 ./ (model.maxx - model.minx);
            else                            cfg.S = 1 ./ model.maxx;    % assumed symmetrical
            end;
        elseif length(cfg.S) ~= model.p,
            error('Scaling vector S is incorrectly sized');
        end;
        
        % Identify portion of sample data corresponding to each discrete action
        Is = cell(M, 1);
        for j = 1:M, Is{j} = findflat(U(:, j), Us); end;

        % If using kdtrees, build them
        if strcmp(cfg.knnmethod, 'kdtree'),
            dispx('Building kd-trees...', cfg.verb, 1);
            tmark = cputime;
            kdtrees = nan(M, 1);
            for j = 1:M, kdtrees(j) = kdtree_build(Xs(:, Is{j})'); end;
            timestat.init_buildtrees = cputime - tmark;
            timestat.init = timestat.init + timestat.init_buildtrees;
            dispx([8 ' done.'], cfg.verb, 1);
            % add pointers to global variable
            KDTREE_POINTERS(end+1:end+M) = kdtrees;
        end;

        % Pre-computation of k nearest neighbors and partial inverses for
        % each sample x discrete action
        
        % Phase 1) search for k nearest neighbors
        dispx('Pre-searching kNNs...', cfg.verb, 1);
        tmark = cputime;            % notice console output below COUNTED IN
        INDLLR = nan(knn, Ns, M);   % indices of NNs
        WLLR = nan(knn, Ns, M);     % initially distances to NNs; converted to weights at next step
        switch cfg.knnmethod,
            case 'kdtree',
                dispx([8 8 8 8 sprintf(' [kdtree] - samples processed: %3d%%', 0)], cfg.verb, 2);
                for is = 1:Ns,
                    for j = 1:M,
                        % find indices & distances of kNNs in the samples for the crt discrete action
                        % note the query point is column vector -- kdtrees package characteristic
                        [ind, dist] = kdtree_k_nearest_neighbors(kdtrees(j), Xps(:, is), knn);
                        % transform into global indices and save; also save distances
                        % note cpp code returns NNs from farthest to closest; we reverse this order
                        INDLLR(knn:-1:1, is, j) = Is{j}(ind);
                        WLLR(knn:-1:1, is, j) = dist;
                    end;
                    if cfg.verb >= 2 && mod(is/Ns, cfg.sampledisp) <= 1/Ns,    % display progress inline
                        disp([8 8 8 8 8 sprintf('%3d%%', fix(is/Ns*100))]);
                    end;
                end; 
            case 'jit',
                % this code runs with discrete-action granularity, for all state samples
                dispx([8 8 8 8 sprintf(' [jit] - disc actions processed: %2d', 0)], cfg.verb, 2);
                for j = 1:M,
                    % find indices & distances for all Xps for discrete action j (notice transposes)
                    [ind, dist] = knnsearchx(Xps', Xs(:, Is{j})', knn, cfg.S);
                    % transform to global indices
                    ind(:) = Is{j}(ind(:));
                    % save indices and distances to arrays (notice transpose)
                    INDLLR(:, :, j) = ind';
                    WLLR(:, :, j) = dist';
                    if cfg.verb >= 2, disp([8 8 8 sprintf('%2d', j)]); end;     % display progress inline
                end;
        end;
        timestat.init_knn = cputime - tmark;
        timestat.init = timestat.init + timestat.init_knn;
        dispx([8 ' done.'], cfg.verb, 1);
        
        % Phase 2) precompute partial inverses
        dispx('Pre-computing pinv...', cfg.verb, 1);
        dispx([8 8 8 8 sprintf(' - samples processed: %3d%%', 0)], cfg.verb, 2);
        tmark = cputime;
        XLLRpinv = nan(p+1, knn, Ns, M);
        xllr = zeros(knn, p);   
        for is = 1:Ns,
            for j = 1:M,
                % transform the distances to the current sample into actual
                % weights, using the weight function
                WLLR(:, is, j) = llrweights(cfg.wfun, WLLR(:, is, j), WLLR(knn, is, j));
                % for each kNN: subtract query sample and multiply by weight
                % notice data points are on the rows of the LHS (xllr)
                % (rely on JIT to optimize this for, rather than using repmat)
                for ik = 1:knn,
                    xllr(ik, :) = WLLR(ik, is, j) .* (Xs(:, INDLLR(ik, is, j)) - Xps(:, is));
                end;
                % last column (weighted) biases, i.e., 1 * weights, i.e., weights
                XLLRpinv(:, :, is, j) = pinv([xllr WLLR(:, is, j)]);
            end;
            if cfg.verb >= 2 && mod(is/Ns, cfg.sampledisp) <= 1/Ns,    % display progress inline
                disp([8 8 8 8 8 sprintf('%3d%%', fix(is/Ns*100))]);
            end;
        end; 
        timestat.init_pinv = cputime - tmark;
        timestat.init = timestat.init + timestat.init_pinv;
        dispx([8 ' done.'], cfg.verb, 1);        

        % init Q-value samples
        Qs = zeros(Ns, 1);
        if cfg.storehistory || cfg.checkosc, Qsh{1} = Qs; end;

        k = 1;                  % init iteration index
        timestat.run = 0;
        
        save(cfg.datafile);
        dispx(['Initialization data saved to [' cfg.datafile '].'], cfg.verb, 1);
    end;        % initialization IF
        
    dispx('Performing LLRFittedQI...', cfg.verb, 0);

    % init visualization config if needed
    if cfg.visualize,
        vcfg = cfg.viscfg;
        vcfg.gview = [];
        vcfg.llrfittedqiter = 1;
        % visualize initial state of the algorithm
        vcfg.ell = 0;
        [figh vcfg.gview] = feval(model.visualizefun, vcfg);
    end;        
    
    % Main Q-iteration loop
    conv = 0;
    Qxp = zeros(M, 1);
    while k <= cfg.maxiter && ~conv,
        
        dispx(sprintf('Iter k=%d, samples processed: %3d%%', k, 0), cfg.verb, 3);
        
        tmark = cputime;
        Qsold = Qs; 
        for is = 1:Ns,
            % find approximate Q-values of all the discrete actions
            for j = 1:M,
                % pick up Q samples for NNs 
                qllr = Qsold(INDLLR(:, is, j));
                % solve linear system to find local regressor; note Q samples are weighted;
                % also note query point is subtracted from data, so we only need the bias term;
                % the entire solution would be: 
                % betallr = XLLRpinv(:, :, is, j) * (qllr .* WLLR(:, is, j));
                biasbetallr = XLLRpinv(p+1, :, is, j) * (qllr .* WLLR(:, is, j));
                % compute regression output for query point;
                if satflag, Qxp(j) = max(min(qllr), min(biasbetallr, max(qllr))); 
                else        Qxp(j) = biasbetallr;
                end;
            end;
            Qs(is) = Rs(is) + cfg.gamma .* max(Qxp);
            % display progress inline -- note this IS counted in
            % runtime -- too much headache to remove it
            if cfg.verb >= 3 && mod(is/Ns, cfg.sampledisp) <= 1/Ns,
                disp([8 8 8 8 8 sprintf('%3d%%', fix(is/Ns*100))]);
            end;
        end;
        
        % update runtime (statistics and user feedback below are not counted)
        timestat.iter_updateQ(k) = cputime - tmark;

        % save Q samples on history if configured
        if cfg.storehistory || cfg.checkosc, Qsh{k+1} = Qs; end;

        % perform convergence, divergence, oscillation tests
        tmark = cputime;
        deltah(k) = max(abs(Qs - Qsold));
        if deltah(k) <= cfg.eps,            
            conv = 1;
        elseif deltah(k) > cfg.maxdelta,    
            conv = -1;
        elseif cfg.checkosc,    % check for oscillations
            per = 2; osc = 0;   % don't test for period = 1, that's taken care of by the simple convergence test
            % try periods up to maxper, as long as oscillation wasn't detected yet
            % and provided that the number of iterations allows for it
            while ~osc && (per <= cfg.maxper) && (per * (cfg.nper+1) + 1 < k+1)
                osc = 1;                % start by assuming oscillating; tests below might falsify that
                for i = 0:per-1,        % i loops through the offsets of the iters that have to be equal
                    for j = 1:cfg.nper, % make sure oscillation happens for at least nper periods
                        osc = osc & all(abs(Qsh{k+1-i} - Qsh{k+1-j*per-i}) <= cfg.epsper);
                        if ~osc, break; end;    % don't test longer periods
                    end;
                    if ~osc, break; end;        % don't test other offets
                end;
                per = per + 1;
            end;
            per = per - 1;                      % revert the last increment
            if osc, conv = 2; end;    % flag oscillation
            % clean up any unnecessary history
            if ~cfg.storehistory && k > cfg.maxper * (cfg.nper+1), Qsh{k - cfg.maxper * (cfg.nper+1)} = []; end;
        end;
        timestat.iter_convtest(k) = cputime - tmark;
        
        % finalize high-verbosity iteration display 
        dispx(sprintf([8 ' done, delta=%g.'], deltah(k)), cfg.verb, 3);
        % visual feedback of algorithm progress
        if ~mod(k, cfg.iterdisp) && cfg.verb < 3, % if >= 3, taken care of above
            dispx(sprintf('Iter k=%d done, delta=%g.', k, deltah(k)), cfg.verb, 2);
        end;
        % visualization
        if cfg.visualize && ~mod(k, cfg.itervis),
            vcfg.ell = k;
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;
        % data backup
        if ~mod(k, cfg.itersave),
            save(cfg.datafile);
            dispx(sprintf('Data at iter k=%d saved to [%s].', k, cfg.datafile), cfg.verb, 2);
        end;
        
        timestat.iter_run(k) = timestat.iter_updateQ(k) + timestat.iter_convtest(k);
        timestat.run = timestat.run + timestat.iter_run(k);
        k = k + 1;
    end;        % WHILE allowed more iterations

    % termination message 
    if conv == 1,       dispx('Convergence detected. Algorithm stopped.', cfg.verb, 0);
    elseif conv == 2,   dispx(['Solution oscillates with period ' num2str(per) '. Algorithm stopped.'], cfg.verb, 0);
    elseif conv == -1,  dispx('Divergence detected! Algorithm stopped.', cfg.verb, 0);
    else                dispx(['maxiter=' num2str(cfg.maxiter) ' exhausted. Algorithm stopped.'], cfg.verb, 0);
    end;

    % finalize visualizer
    if cfg.visualize,
         % make sure last iteration is visualized
        if vcfg.ell < k - 1,    vcfg.ell = k - 1;       % show last iteration
        else                    vcfg.llrfittedqiter = 0;    % don't show anything
        end;
        vcfg.finalize = 1;
        [figh vcfg.gview] = feval(model.visualizefun, vcfg);
    end;

    % create an appropriate approximator (to allow generic Q and h
    % computation)
    acfg = struct;
    acfg.U = cfg.U;
    acfg.knnmethod = cfg.knnmethod; acfg.knn = cfg.knn; 
    acfg.S = cfg.S;
    acfg.wfun = cfg.wfun;
    acfg.linsolvemethod = cfg.linsolvemethod;
    acfg.extrap = cfg.extrap;
    acfg.Xs = Xs; acfg.Is = Is; acfg.Qs = Qs;
    if strcmp(cfg.knnmethod, 'kdtree'), cfg.kdtrees = kdtrees; end;
    approx = llrdisc(model, acfg); clear acfg;
    
    % deallocate tree memory
    if strcmp(cfg.knnmethod, 'kdtree'),
        for j = 1:M, kdtree_delete(kdtrees(j)); end;
    end;
    
    % use this line to cleanup history, if only used to check for oscillation
    % for now, keep it in case we need to resume later
    % if cfg.checkosc && ~cfg.storehistory, clear Qsh; end;
end;


% Options below might create figures, init figure handles array
figh = [];
if isempty(cfg.savefig), cfg.savefig = cfg.datafile; end;

% ==== REPLAY POLICY -- Including intermediate plots and trajectory saving due to high comp cost ====
if cfg.replay,

    % initial state
    if ~isempty(cfg.x0), 
        if ischar(cfg.x0) && ~isempty(cfg.problem),
            % use the model function to get a string-named initial state
            x0 = feval(cfg.problem, 'x0', cfg.x0);
        else
            x0 = cfg.x0(:);         % specified initial state
        end;
    else    % try getting default initial state
        try         x0 = feval(cfg.problem, 'x0');
        catch       x0 = zeros(model.p, 1);    % zeros
        end;
    end;         
    dispx(['Controlling from x0=' num2str(reshape(x0, 1, [])) ], cfg.verb, 0);

    % init history
    t = 0 : model.Ts : cfg.tend;
    Ns = length(t)-1;       % number of samples / time instances at which control is applied
    x = zeros(model.p, length(t)); x(:, 1) = x0;
    u = zeros(model.q, length(t)); u(:, end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    % steps loop
    th = NaN;   % time elapsed while computing action for one state
    for k = 1:Ns,
        dispx(sprintf('k=%d/%d. th=%f', k, Ns, th), cfg.verb, 5);
        u(:, k) = approx.h(approx, Qs, x(:, k));
        [x(:, k+1) r(k+1) terminal] = feval(model.fun, model, x(:, k), u(:, k));
        if terminal, Ns = k; break; end;
    end;
 
    % create & optionally save history
    hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1);
    hist.R = discreturn(cfg, hist.r, Ns, terminal);
    % plot history & optionally save figures
    if ~cfg.silent,
        if isfield(model, 'plotfun'),
            mpfigh = feval(model.plotfun, hist);
            if ~isempty(mpfigh), figh(end+1) = mpfigh; end;
        end;
        % if no plot function or plot function did not create a figure
        if ~isfield(model, 'plotfun') || isempty(mpfigh),
            figh(end+1) = plothistory(hist);
        end;
        setfigprop(cfg);
        % save if requested
        saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
    end;
end;        % IF replay

% ==== STATISTICS ====
if (cfg.evol || cfg.sol) && ~cfg.silent,
    % grayscale styles
    sty.cm = gray(128); sty.cm = sty.cm(33:end-5, :);  % w/o strong blacks and whites
    % labels
    labels.iter = 'Iteration'; 
    labels.h = 'h(x)'; labels.V = 'V(x)'; 
    labels.delta = {'delta'}; 
    figh = [];
end;

if cfg.sol && ~cfg.silent,
    figh(end+1) = figurex([1100 450]);
    K = find(~isnan(deltah), 1, 'last') - 1;
    subplot(3, 2, [1 3]); cla;
    pcfg = struct; pcfg.npoints = 30; pcfg.slice = cfg.slice;
    approx.plotv(approx, Qs, pcfg); title(labels.V);
    subplot(3, 2, [2 4 6]); cla;
    pcfg = struct; pcfg.npoints = 70; pcfg.slice = cfg.slice;
    approx.ploth(approx, Qs, pcfg); title(labels.h);
    subplot(3, 2, 5); cla;
    semilogy(1:K, deltah(1:K), 'k'); grid on; xlabel(labels.iter); ylabel(labels.delta{:});
    setfigprop(cfg);
    % save if requested
    saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
end;

if cfg.evol && ~cfg.silent,
    % TODO
end;


% ==== BACKUP DATA ====
if cfg.run || cfg.resume, 
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir; 
        if any(datadir(end) == '\/'), datadir = datadir(1:end-1); end;
    end;
    % if default file name is used, generate a unique file name
    if strcmp(cfg.datafile, CFG.datafile),
        c = fix(clock);
        for i = 1:length(c),
            cfg.datafile = [cfg.datafile '-' num2str(c(i), '%02d')];
        end;
        if exist([CFG.datafile '.mat'], 'file'), delete([CFG.datafile '.mat']); end;
    elseif ~strcmp(datadir, pwd)
        % need to save in a different directory, so delete anyway
        if exist([cfg.datafile '.mat'], 'file'), delete([cfg.datafile '.mat']); end;
    end;    % otherwise just overwrite
    cfg.datafile = [datadir '/' cfg.datafile];
    save(cfg.datafile);
    dispx(['Final data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;

% -----------------------------------------------
% set output
if cfg.run || cfg.resume,       % output optimal solution
    % varargout = {approx};
elseif cfg.replay,              % output history and possibly figure handles
    varargout = {hist, figh};
elseif cfg.evol || cfg.sol,               % output fig handles (replay takes precedence)
    varargout = {figh};
end;

end
% lspi() RETURNING varargout =================================================================
