function varargout = rbfqi(cfg)
% Approximate offline, model-based Q-iteration with RBF approximation
%   VARARGOUT = RBFQI(CFG)
% This is currently a wrapper for cerbfqiter. As a result, some functionality such as 
% resuming and the reusing approximator structure, is not provided.
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see he str2cfg
% Outputs:
%   THETASTAR, QISTATS	- if run or resume mode. THETASTAR is the optimal parameter vector,
%           QISTATS are the Q-iteration statistics
%   HIST, FIGH          - in replay mode. HIST is the replay history. 
%           FIGH contains figure handles if figures were created 
%           and not closed; and is an empty matrix if all the figures were closed.
%

% Note does not support model_params !!

if nargin < 1, cfg = struct(); end;

% -----------------------------------------------
% Process configuration structure

% default config
% script config
CFG.run = 0;                        % run learning
CFG.resume = 0;                     % resume learning -- no resume supported currently
CFG.replay = 0;                     % replay learned policy
CFG.problem = '';                   % what problem to solve
CFG.loadapprox = '';                % try loading approximator data from file
CFG.datafile = 'rbfqidata';          % save data to file
% learning config
CFG.c = [];                      	% rbf centers and radii; defined or provided by problem
CFG.rad = [];
CFG.ugrids = [];                    % action quantization; same comments as above
CFG.gamma = 0.98;                   % discount factor
CFG.eps = .01;                      % threshold for convergence
CFG.maxiter = 300;                  % max number of iterations for Q-iteration
CFG.usesparse = 1;                  % whether to use sparse matrices in the model
CFG.cutoff = 1e-7;                  % cutoff value; all RBF activations below this value will be considered 0
                                    % ignored when usesparse = 0
% replay config
CFG.returnprecision = [];           % replay s.t. return can be evaluated with this precision
CFG.replaymodelfile = '';           % use model from a different experiment (data file)
CFG.interph = 0;                    % interpolated (averaged) policy
CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 30;                      % end time for replay
% stats & figure output config
CFG.truncatehist = [];              % whether history should be truncated at this time value (in seconds)
CFG.plottarget = 'screen';          % 'screen', '', 'latex', or 'beamer'. If 'screen' figures will not be closed
CFG.savetheta = 0;                  % save param history in stats
CFG.savedir = '';
CFG.savefig = '';
CFG = setfigprop(CFG, 'addfields');
% residual options
CFG.bellmanres = 0;
CFG.X0 = [];
% display config
CFG.verb = 5;                       % verbosity: the higher, the more detailed the messages displayed
CFG.noplot = 0;                     % whether to suppress figure plots
CFG.silent = 0;                     % suppress all output
CFG.initstepdisp = .1;              % feedback every 10% of MDP init
CFG.iterdisp = 10;                  % feedback after every 10 iterations
CFG.itersave = 25;                  % save after each 25 iterations

% Early defaults (initialized before calling problem defaults)
ECFG.rbf_params = {};             % extra parameters for problem calling in 'fuzzy' mode
ECFG.model_params = {};             % extra parameters for problem calling in 'model' mode

% List of fields that define the problem and therefore may NOT be overwritten on load
KEEPFIELDS = {'gamma', 'eps', 'maxiter'};

% If caller provided string, parse it into a structure
if ischar(cfg),
    cfg = str2cfg(cfg, fieldnames(CFG));
end;
% Try installing problem defaults whenever caller does not specify values
% Problem must be specified for this to work
try     
    if ~isempty(cfg.problem), cfg = checkparams(cfg, feval(cfg.problem, 'rbf'), cfg.rbf_params{:}); end;
catch   % do nothing
end;
% Install function defaults for everything else
cfg = checkparams(cfg, CFG);
% Process configuration dependencies
if cfg.silent, cfg.verb = -Inf; cfg.noplot = 1; end;
% Echo config
dispx('RBF Q-iteration called with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1);

% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

cfg.noinit = (cfg.resume || cfg.replay || cfg.bellmanres) && exist([cfg.datafile '.mat'], 'file');

if cfg.noinit,        % load data file, making sure that cfg is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile);
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Current data loaded from [' cfg.datafile '].'], cfg.verb, 1);
end;

if ~cfg.noinit,
    % -----------------------------------------------
    % Setup model
    model = feval(cfg.problem, 'model', cfg.model_params{:});

    % -----------------------------------------------
    % Setup approximator structure, if not loaded; creates c, rad, U, DIMS
    % ensure data was supplied
    checkparams(cfg, [], {'c', 'rad', 'ugrids'});
    c = cfg.c; rad = cfg.rad; U = cfg.ugrids;
    if iscell(c), c = flat(c); end;     % a grid of center coords was given
    if any(size(c) ~= size(rad)),
        if isscalar(rad),       % uniform radii for all RBFs, all dimensions
            rad = rad + 0*c;
        elseif isvector(rad),   % uniform radii across each dimension
            % assuming the vector size is = model.p
            rad = repmat(rad(:), 1, size(c, 2));
        else
            error('Invalid RBF radius specification');
        end;
    else
        % do nothing, rad is explicitly given
    end;
        
    % flatten U
    U = flat(U);
    
    % # of parameters
    DIMS.p = model.p; DIMS.q = model.q;
    DIMS.N = size(c, 2);
    DIMS.M = size(U, 2);
    
    DIMS.dimx = []; DIMS.dimu = [];             % dimensions of grids
    for p = 1:DIMS.p, DIMS.dimx(end+1) = length(cfg.xgrids{p}); end;
    for q = 1:DIMS.q, DIMS.dimu(end+1) = length(cfg.ugrids{q}); end;

    % shorthand dimension variables
    N = DIMS.N; M = DIMS.M;
end;


% -----------------------------------------------
% Q-iteration
if cfg.run || cfg.resume,

    if ~cfg.resume,
%         % determine if previously computed data can be used
%         if ~isempty(cfg.loadapprox),
%             vars = {'cfg', 'model', 'c', 'rad', 'Ainv', 'F', 'R'};
%             loadwithprefix(cfg.loadapprox, 'app', vars);
%             % test if approximator is good
%             useapprox = (cfg.sparse == appcfg.sparse) && (cfg.cutoff == appcfg.cutoff) ...
%                 && isequal(model, appmodel) && all(all(c == appc)) && all(all(rad == crad));
%         else useapprox = 0;
%         end;
%         
%         if useapprox,
%             dispx(['Using approximator data from file [' cfg.loadapprox ']'], cfg.verb, 1);
%             Ainv = appAinv;
%             F = appF;
%             R = appR;
%             clear appAinv appF appR;
%         else
            dispx('Computing MDP and activation data...', cfg.verb, 1);
            tmark = cputime;
            % activation of each center
            if cfg.usesparse, A = sparse(N, N); else A = zeros(N, N); end;
            % for each x and u, activation vector of f(x, u)
            if cfg.usesparse, F = sparse(N * M, N); else F = zeros(N * M, N); end;
            % reward rho(x, u)
            R = zeros(N, M);

            if cfg.usesparse,
                prog = cfg.initstepdisp;           
                cutoff = cfg.cutoff;
                for i = 1:N,
                    x = c(:, i);
                    [ind, a] = nrbf_cutoff(x, N, c, rad, cutoff); 
                    A(i, ind) = a;
                    for j = 1:M,
                        [xplus R(i, j)] = feval(model.fun, model, x, U(:, j));
                        [ind, a] = nrbf_cutoff(xplus, N, c, rad, cutoff);
                        F(i+(j-1)*N, ind) = a;
                    end;            % FOR over action discretization
                    if i/DIMS.N > prog,     % progress feedback
                        dispx([num2str(prog * 100) '% completed...'], cfg.verb, 2);
                        prog = prog + cfg.initstepdisp;
                    end;
                end;                % FOR over state samples
            else
                prog = cfg.initstepdisp;           
                % iterate over (x, u) in (centers, discrete actions) and populate A, F, R structures
                for i = 1:N,
                    x = c(:, i);
                    % compute activation vector
                    A(i, :) = nrbf(x, N, c, rad);    
                    for j = 1:M,
                        % apply f, rho, and store results
                        [xplus R(i, j)] = feval(model.fun, model, x, U(:, j));
                        F(i+(j-1)*N, :) = nrbf(xplus, N, c, rad);
                    end;            % FOR over action discretization

                    if i/DIMS.N > prog,     % progress feedback
                        dispx([num2str(prog * 100) '% completed...'], cfg.verb, 2);
                        prog = prog + cfg.initstepdisp;
                    end;
                end;                % FOR over state samples
            end;

            % precompute the inverse of A
            dispx('Computing the inverse of the activation matrix...', cfg.verb, 2);
            Ainv = inv(A);

            % init time includes time to compute inverse
            qistats.tinit = cputime - tmark;
%         end;

        dispx('Performing linear Q-value iteration...', cfg.verb, 1);

        % initialize linear parameter matrix
        if nargin < 5,  theta = zeros(N, M);
        else            theta = theta0;
        end;

        k = 1;
        delta = nan(1, cfg.maxiter);
        conv = 0; div = 0; t = cputime;
    end;        % IF init required
        
    while ~conv && ~div && k <= cfg.maxiter,       % main loop

        tmark = cputime;
        thetaold = theta;
        % apply least-squares update
        theta = Ainv * (R + cfg.gamma * reshape(max(F * theta, [], 2), N, M));

        % compute infty-norm of vector representation of theta difference
        delta(k) = max(max(abs(theta - thetaold)));        
        conv = delta(k) <= cfg.eps;
        div = k > 10 && all(diff(delta(k-10:k)) > 0);

        k = k + 1;
        t = t + (cputime - tmark);

        if ~mod(k, cfg.iterdisp),       % visual feedback of algorithm progress
            dispx(['k=' num2str(k) ' iteration completed, last deltas=' num2str(delta(max(1, k-5):k-1))], cfg.verb, 2);
        end;
        if ~mod(k, cfg.itersave),       % data backup
            save(cfg.datafile);
            dispx(['Intermediary data at k=' num2str(k) ' saved to [' cfg.datafile '].'], cfg.verb, 2);
        end;
    end;        % while not converged and allowed more iterations

    if conv,	flag = 0; dispx(['Convergence detected after ' num2str(k-1) ' iterations. Algorithm stopped.'], cfg.verb, 1);
    elseif div, flag =-1; dispx(['Divergence detected after ' num2str(k-1) ' iterations. Algorithm stopped.'], cfg.verb, 1);
    else        flag = 1; dispx(['maxiter=' num2str(cfg.maxiter) ' exhausted. Algorithm stopped'], cfg.verb, 1);
    end;
        
%        % Call to rbfqiter for the actual Q-iteration algorithm
%        innercfg.gamma = cfg.gamma; innercfg.maxiter = cfg.maxiter; innercfg.eps = cfg.eps;
%        innercfg.N = DIMS.N; innercfg.M = DIMS.M; innercfg.U = U;
%        innercfg.verb = cfg.verb;
%        [thetastar, conv, delta] =  rbfqiter(c, rad, model, innercfg);   % theta0 = default, 0
       
   % also save vars compatible with CERBFQI/PS replays, data plots etc.
   thetastar = theta;
   phistar = [c; rad];
   qistats.delta = delta;
   qistats.trun = t;
end;

% -----------------------------------------------
% Replay
if cfg.replay,

    if ~isempty(cfg.replaymodelfile),
        loadwithprefix(cfg.replaymodelfile, 'rep_', 'model');
%         % superficially check the model, revert to the default experiment model if check fails
%         rmok = (rep_model.Ts == model.Ts) && (rep_model.p == model.p) && (rep_model.q == model.q);
%         if ~rmok, 
%             warning('APPROXRL:generic', 'Indicated model does not match model used for learning. Reverting');
%             rep_model = model; 
%         end;
    else
        rep_model = model;
	end;
    
    % compute locally optimal policy (local optimal discrete action for each basis
    % function)
    if cfg.interph
        [Qstar ui] = max(theta, [], 2); clear Qstar;
        hstar = U(:, ui)';
    end;

    % initial state
    if ~isempty(cfg.x0), x0 = cfg.x0(:);      % specified initial state
    else x0 = zeros(model.p, 1); end;         % zeros
    dispx(['Controlling from x0=' num2str(reshape(x0, 1, [])) ], cfg.verb, 0);

    if ~isempty(cfg.returnprecision),
        K = ceil(log(cfg.returnprecision * (1-cfg.gamma) / rep_model.maxr) / log(cfg.gamma));
        cfg.tend = K * model.Ts;
    end;
    % init history
    t = 0 : model.Ts : cfg.tend;
    Ns = length(t)-1;       % number of samples / time instances at which control is applied
    x = zeros(model.p, length(t)); x(:, 1) = x0;
    u = zeros(model.q, length(t)); u(:, end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    % steps loop
    for k = 1:Ns,
        % compute optimal action
        if cfg.interph,     % either interpolated
            u(:, k) = nrbf(x(:, k), DIMS.N, c, rad) * hstar;
        else                % or crisp (ties broken by choosing first best action)
            Qa = nrbf(x(:, k), DIMS.N, c, rad) * theta;
            u(:, k) = U(:, find(Qa == max(Qa), 1));
        end;
        % apply to system
        [x(:, k+1) r(k+1) terminal] = feval(rep_model.fun, rep_model, x(:, k), u(:, k));
        if terminal, Ns = k; break; end;      % entered terminal state
    end;
    R = discreturn(cfg, r, Ns, terminal);
 
    % plot history & optionally save figures
    hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1); hist.R = R(1:Ns+1);
    if ~isempty(cfg.truncatehist), hist = truncatehist(hist, cfg.truncatehist); end;
    if ~cfg.noplot,
        toscreen = strcmp(cfg.plottarget, 'screen') || isempty(cfg.plottarget);
        figh = plothistory(hist);
        setfigprop(cfg);
        saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
        if isfield(model, 'plotfun'),
            mpfigh = feval(model.plotfun, hist);
            if ~isempty(mpfigh), figh(end+1) = mpfigh; end;
        end;
    end;
        
end;        % IF replay

if cfg.bellmanres,
    dispx('Computing Bellman residual...', cfg.verb, 2);
    % Init transition and rewards table for set of interesting states
    N0 = size(cfg.X0, 2);
    beta = ones(N0, DIMS.M) ./ (N0 * DIMS.M);
    br = 0;
    for i = 1:N0,
        for j = 1:DIMS.M,
            [xplus rplus] = feval(model.fun, model, cfg.X0(:, i), U(:, j));
            br = br + beta(i, j) * ...                                      % weight
                ( nrbf(cfg.X0(:, i), DIMS.N, c, rad) * thetastar(:, j) ...         % (lhs 
                  - (rplus + cfg.gamma * max(nrbf(xplus, DIMS.N, c, rad) * thetastar)) )^2;  % - rhs)^2
        end;
    end;
    % display residual
%     br
end;


% -----------------------------------------------
% Backup data

if cfg.run || cfg.resume, 
    % save into the same directory as the problem
    savedir = fileparts(which(cfg.problem));
    % if default file name is used, generate a unique file name
    if strcmp(cfg.datafile, CFG.datafile),
        c = fix(clock);
        for i = 1:length(c),
            cfg.datafile = [cfg.datafile '-' num2str(c(i), '%02d')];
        end;
        if exist([CFG.datafile '.mat'], 'file'),
            delete([CFG.datafile '.mat']);
        end;
    elseif ~strcmp(savedir, pwd)
        % need to save in other directory so delete anyway
        if exist([CFG.datafile '.mat'], 'file'),
            delete([CFG.datafile '.mat']);
        end;
    end;    % otherwise just overwrite
    cfg.datafile = [savedir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['RBF Q-iteration finished. Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;

% -----------------------------------------------
% set output
if cfg.run || cfg.resume,       % output Q-iteration statistics
    varargout = {theta, qistats};
elseif cfg.replay               % output history and possibly figure handles
    fig = ~cfg.noplot && (strcmp(cfg.plottarget, 'screen') || isempty(cfg.plottarget));
    if fig,         varargout = {hist, figh};
    else            varargout = {hist, []};
    end;
elseif cfg.bellmanres
    varargout = {br};
end;


% END fuzzyqi() RETURNING varargout =================================================================

