function varargout = rtapproxqlearn(cfg)
% Real-time approximate Q-learning with a linear parametrization and gradient-based updates
%   VARARGOUT = RTAPPROXQLEARN(CFG)
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see he str2cfg
% Outputs:
%   THETA           - in run mode. Found parameter vector.
%   HIST, FIGH      - in replay mode -- NOT IMPLEMENTED. HIST is the replay history.
%           FIGH contains figure handles if figures were created 
%           and not closed; and is an empty matrix if all the figures were closed.

if nargin < 1, cfg = struct; end;

% -----------------------------------------------
% Process configuration structure

% default config
% script config
CFG.problem = '';                   % what problem to solve
CFG.system_params = {};             % extra parameters for problem calling in 'sys' mode
CFG.run = 0;                        % run learning
CFG.replay = 0;                     % replay learned policy
CFG.datadir = '';                   % data dir
CFG.datafile = '';                  % save data to file
CFG.approx = [];                    % approximator object (or config for approx)
CFG.gamma = [];                     % discount factor
CFG.alpha = 0.1;                    % learning rate
CFG.explor = 1;                     % exploration rate
CFG.explordecay = 0.99;             % exploration decay (per trial)
CFG.lambda = 0;                     % eligibility trace parameter
CFG.eps = 0.01;                     % threshold for convergence
CFG.maxtime = 100;                  % max length of learning experiment (in seconds)
CFG.trialtime = 10;                 % max length of a trial (in seconds)
CFG.reset = 'rand';                 % trial reset
% experience replay configuration
CFG.er_type = 'none';               % 'none', 'rand'=rand trial, rand transition, 'randback' rand trial, backward transition
CFG.er_N = 500;                     % how many transitions to replay after every trial
CFG.randseed = [];                  % set the random seed (for repeatable results)
% replay, etc. config
% CFG.x0 = [];                        % initial state for replay (otherwise the problem default or zeros)
CFG.tend = 10;                      % end time for replay
% output config
% note visualization is not implemented (unlikely to be needed for a real sys)
CFG.trialsave = 50;                 % save once this many trials
CFG.plottarget = 'screen';          % 'screen', '', 'latex', or 'beamer'. If 'screen' figures will not be closed
CFG.savedir = '';
CFG.savefig = '';
% display config
CFG.silent = 0;                     % run in silent mode (e.g., during replay)
CFG.verb = 5;                       % verbosity: the higher, the more detailed the messages displayed
CFG.trialdisp = 10;                 % display console feedback once this many trials

% List of fields that define the problem and therefore may NOT be overwritten on load
KEEPFIELDS = {'problem', 'gamma'};

% Install function defaults for everything else
cfg = parseconfig(cfg, CFG);

cfg.init = cfg.run;

% If running, check presence of data file, warn on possible overwrite
if cfg.run && exist([cfg.datafile '.mat'], 'file'),
    reply = input(['File [' cfg.datafile '.mat] already exists (possible overwrite). Continue? Y/N [N]: '], 's');
    if isempty(reply) || reply == 'N', return; end;
end;       

if ~cfg.init,        % load data file, making sure that cfg is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile);
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
    approx = revise_approx(approx);
end;

% Echo config
dispx('Real-time approx Q-learning called with the config:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1);

% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% -----------------------------------------------
% Create system and approximator, find dimensionality data
if cfg.init,
    sys = feval(cfg.problem, 'system', cfg.system_params{:});
    Ts = sys.Ts;
    % heuristic check if we already have an approx object
    if isstruct(cfg.approx) && isfield(cfg.approx, 'N') && isfield(cfg.approx, 'phi') ...
            && isfield(cfg.approx, 'q') && isfield(cfg.approx, 'h'),
            % nothing to do
            dispx('Approx object supplied', cfg.verb, 3);
    else    % create approximator using cfg.approx as the config
        cfg.approx = create_approx(sys, cfg.approx);
    end;
    approx = cfg.approx;
    N = approx.N; p = sys.p; q = sys.q;
    if approx.disc,        % initialize action grids and their size
        discU = cfg.approx.U;
        flatU = flat(discU); nU = size(flatU, 2);
        % discUn = zeros(length(discU)); for i=1:length(discU), discUn(i) = length(discU{i}); end;        
    end;
    if ~approx.extendedsyntax,
        error('Approximator that supports extended syntax required');
    end;
end;

% -----------------------------------------------
% Learning
if cfg.run,
    
    % derive Ktrial and maxtrials
    maxtrials = cfg.maxtime / cfg.trialtime;
    Ktrial = cfg.trialtime / sys.Ts;
    
    % init Q-function, elig trace, histories etc.
    X = NaN(p, Ktrial+1, maxtrials);
    U = NaN(q, Ktrial+1, maxtrials);
    R = NaN(1, Ktrial+1, maxtrials);
    theta = zeros(N, 1);
    E = zeros(N, 1);                    % eligibility trace
    thetah = cell(maxtrials+1, 1);  
    thetah{1} = theta;                  % save theta_0 on the stats
    deltah = NaN(maxtrials+1, 1);
    
    k = 1;                              % total step counter
    trial = 1;                          % trial counter
    timestat.run = 0;
    explor = cfg.explor;
    alpha = cfg.alpha;
    
    % init system
    sys = feval(cfg.problem, 'init', sys);
    
    if ~isempty(cfg.randseed), rand('twister', cfg.randseed); end;
    
    % -----------------------------------------------
    % Perform Q-iteration
    if cfg.lambda > 0, dispx('Performing real-time approx Q(lambda)-learning...', cfg.verb, 0);
    else dispx('Performing real-time approx Q-learning...', cfg.verb, 0);
    end;

    t = cputime;
    conv = 0;
    while trial <= maxtrials && ~conv,       % main loop

        % TODO FIX: the first state should be obtained from the reset procedure
        
        % initialize trial
        % reset to initial state
        feval(cfg.problem, 'reset', sys, cfg.reset);
        % step counter, terminal flag
        ktrial = 1; % term = 0;
        % reset elig trace
        E(:) = 0;

        % dispx(sprintf('===> Trial %d started // k=%d ktrial=%d trial=%d t=%f', trial, k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);

        timestat.run = timestat.run + (cputime - t);
        t = cputime;    % start measuring execution time
        tic;            % and discrete-time loop
        
        % note we iterate one extra step, to allow the Q-learning update for the last transition 
        % (to sample # Ktrial+1) to run
        while ktrial <= Ktrial+1, % && ~term,        
            
            % dispx(sprintf('===> Measuring x // k=%d ktrial=%d trial=%d t=%f', k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);
            % measure state
            X(:, ktrial, trial) = feval(cfg.problem, 'readx', sys);

            % === BEGIN TIME-CRITICAL -- action choice code
            if ktrial <= Ktrial,      % must not apply action after last sample
                % choose action according to current state
                exploring = rand < explor;
                % choose control action
                if exploring,  % explore, choose uniform random action
                    if approx.disc,    % flatU is a set of discrete actions, each action on a column
                        ucandidate = flatU(:, unidrnd(nU));
                    else        % U is continuous, assumed symmetrical w/ bounds in maxu
                        ucandidate = (2*rand(q, 1) - 1) .* sys.maxu;
                    end;
                else            % exploit using policy greedy in the current theta
                    ucandidate = approx.h(approx, theta, X(:, ktrial, trial));
                end;
                
                % dispx(sprintf('===> Applying u // k=%d ktrial=%d trial=%d t=%f', k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);
                % apply the action
                % note the system may return a modified (e.g., filtered) action
                % if this happens, discrete-action approximators may be broken!
                U(:, ktrial, trial) = feval(cfg.problem, 'writeu', sys, ucandidate);
            else
                % trial has ended, make the system idle
                feval(cfg.problem, 'idle', sys);
            end;
            % === END TIME-CRITICAL
            
            % from now on, we have the whole sample time -- but we should not exceed it
            
            % if we have a transition to work on, perform a learning update
            %               -- always working on the previous transition sample!
            if ktrial > 1,
                % dispx(sprintf('===> Processing transition %d -> %d // k=%d ktrial=%d trial=%d t=%f', ktrial-1, ktrial, k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);                

                % apply system function to compute reward & check if transition was terminal
                [R(1, ktrial, trial), term] = ...
                    feval(sys.fun, sys, X(:, ktrial-1, trial), U(:, ktrial-1, trial), X(:, ktrial, trial));
                if term, error('APPROXQLEARN: terminal state support not implemented'); end;

                % process elig trace if using it
                if cfg.lambda > 0,
                    if exploring,   E(:) = 0;
                    else            E = (cfg.gamma * cfg.lambda) .* E;
                    end;
                end;

                % compute temporal difference and update Q-function parameters
                [Qxu, phixu] = approx.q(approx, theta, X(:, ktrial-1, trial), U(:, ktrial-1, trial));
                [up, Qstarxpup] = approx.h(approx, theta, X(:, ktrial, trial));
                tempdiff = R(1, ktrial, trial) + cfg.gamma * Qstarxpup - Qxu;
                if cfg.lambda > 0,  % use elig trace
                    E = E + phixu;
                    theta = theta + (alpha * tempdiff) .* E;
                else                % no trace
                    theta = theta + (alpha * tempdiff) .* phixu;
                end;

                % (in the future, if alpha must be annealed, this should be done here)
            end;
            
            % update stats
            timestat.run = timestat.run + (cputime - t);
            
            ktrial = ktrial + 1;
            
            % step done; unless extra sample, increment k and wait out the rest of the sample period
            if ktrial <= Ktrial+1, 
                k = k + 1;
                if toc > sys.Ts, disp('WARNING! Took longer than Ts'); end;
                while toc < sys.Ts, end; tic; 
            end;

            t = cputime;        % note we ignore the wait time above
        end;        % steps while
      
        % experience replay if configured to do so
        % note this code must be overhauled if used with episodic tasks (where trajectories can be
        % shorter than Ktrial)
        % if alpha is time-varying, need to consider whether it must be altered here
        if ~strcmp(cfg.er_type, 'none'),
            dispx(sprintf('Replaying %d transitions, method=[%s]', cfg.er_N, cfg.er_type), cfg.verb, 3);
            er_trial = []; er_ktrial = [];
            for i = 1:cfg.er_N,
                % choose trajectory and transition
                switch cfg.er_type,
                    case 'rand',        % random trial, random transition
                        er_trial = unidrnd(trial); er_ktrial = undidrnd(Ktrial);
                    case 'randback',    % random trial, then replay transitions in backward order
                        if isempty(er_trial) || er_ktrial == 1,
                            % choose a new trial
                            er_trial = unidrnd(trial); er_ktrial = Ktrial;
                        else
                            % decrement transition
                            er_ktrial = er_ktrial - 1;
                        end;
                    case 'back',        % replay trials and transitions in backward order
                        if isempty(er_trial),
                            % initialize
                            er_trial = trial; er_ktrial = Ktrial;
                        elseif er_ktrial == 1,
                            % last transition, decrement trial (with rollover from 0 to last trial)
                            if er_trial == 1,   er_trial = trial;
                            else                er_trial = er_trial - 1;
                            end;
                        else
                            % decrement transition
                            er_ktrial = er_ktrial - 1;
                        end;
                    otherwise
                        errror('Unknown ER type [%s]', cfg.er_type);
                end;
                % replay (update Q-function with) the chosen transition
                [Qxu, phixu] = approx.q(approx, theta, X(:, er_ktrial, er_trial), U(:, er_ktrial, er_trial));
                [up, Qstarxpup] = approx.h(approx, theta, X(:, er_ktrial+1, er_trial));
                tempdiff = R(1, er_ktrial+1, er_trial) + cfg.gamma * Qstarxpup - Qxu;
                theta = theta + (alpha * tempdiff) .* phixu;
            end;
        end;
            
        % store Q-function on history
        thetah{trial+1} = theta;
        % compute max absolute difference
        deltah(trial+1) = max(max(abs(theta - thetah{trial})));
        conv = deltah(trial+1) < cfg.eps;

        % update execution time -- before starting to output to console etc.
        timestat.run = timestat.run + (cputime - t);
        
        % console feedback of algorithm progress
        if ~mod(trial, cfg.trialdisp), 
            dispx(sprintf('Trial #%d completed, delta=%.3f', trial, deltah(trial+1)), cfg.verb, 2);
        end;
        % data backup
        if ~mod(trial, cfg.trialsave),
            save(cfg.datafile);
            dispx(sprintf('Data after trial #%d saved to [%s].', trial, cfg.datafile), cfg.verb, 1);
        end;
        
        % start counting time again, increment trial counter
        t = cputime;
        trial = trial + 1;
        % anneal exploration
        explor = explor * cfg.explordecay;
    end;        % trial while

    if conv,	dispx('Convergence detected. Algorithm stopped.', cfg.verb, 0);
    else        dispx(sprintf('maxtime=%f exhausted. Algorithm stopped.', cfg.maxtime), cfg.verb, 0);
    end;
        
    % close system
    sys = feval(sys.problem, 'close', sys);
    
    % output parameter vector
    varargout = {theta};
end;


% Options below might create figures, init figure handles array
figh = [];

% Replay using learned policy
if cfg.replay,

    % init system
    sys = feval(cfg.problem, 'init', sys);        
    % reset to initial state
    x0 = feval(cfg.problem, 'reset', sys, cfg.reset);
    dispx(['Controlling from x0=' num2str(reshape(x0, 1, [])) ], cfg.verb, 0);

    % init history
    t = 0 : sys.Ts : cfg.tend;
    Ns = length(t)-1;       % number of samples / time instances at which control is applied
    x = zeros(sys.p, length(t)); x(:, 1) = x0;
    u = zeros(sys.q, length(t)); u(:, end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    % steps loop
    tic;
    for k = 1:Ns+1,
        x(:, k) = feval(cfg.problem, 'readx', sys);
        % time-critical section
        if k <= Ns,     % need to apply control
            ucandidate = approx.h(approx, theta, x(:, k));
            u(:, k) = feval(cfg.problem, 'writeu', sys, ucandidate);
        end;
        % now we have the rest of the sample time
        if k > 1,       % we have a transition to process
            [r(1, k), term] = feval(sys.fun, sys, x(:, k-1), u(:, k-1), x(:, k));
        end;
        if term, error('Terminal state support not implemented'); end;
        while toc < Ts, end; tic;
    end;
    % close system
    sys = feval(cfg.problem, 'close', sys);        

    % plot history & optionally save figures
    history.t = t(1:Ns+1); history.x = x(:, 1:Ns+1); history.u = u(:, 1:Ns+1); history.r = r(1:Ns+1);
    history.R = discreturn(cfg, history.r, Ns, term);
    if ~cfg.silent,
        if isfield(sys, 'plotfun'),
            mpfigh = feval(model.plotfun, history);
            if ~isempty(mpfigh), figh = mpfigh; end;
        else
            figh = plothistory(history);
        end;
        setfigprop(cfg);
        % save if requested
        saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
    end;
end;        % IF replay


% -----------------------------------------------
% Backup data

if cfg.run && ~isempty(cfg.datafile),
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir;
        if (datadir(end) == '\'), datadir = datadir(1:end-1); end;
    end;
    cfg.datafile = [datadir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;


% END qi() RETURNING varargout =================================================================
