function out = cleanrob_problem(mode, varargin)
% Cleaning robot problem. Conforms to specs of sample_problem.

switch mode,

    % Offer basic info about the model (without creating the actual model)
    case 'info',    
        info.id = 'cleanrob';               % an identifier
        info.problem = @cleanrob_problem;   % handle to model function
        info.det = [];                      % can be either deterministic or stochastic
        info.p = 1;                         % # of state variables
        info.q = 1;                         % # of control variables
        out = info;
        
    case 'shortid', out = 'cleanrob';

    % create the model
    case 'model',
        
        model.p = 1;                        % number of states
        model.q = 1;                        % number of control inputs (actions)
        model.nw = 0;                       % number of noise variables
        model.Ts = 1;                       % no sample time per se, just a discrete time index
        model.fun = 'cleanrob_mdp';         % MDP function
        % visualization function
        model.visualizefun = 'cleanrob_visualize';
        model.userefmodel = 0;
        
        % configurable properties
        CFG.gamma = 0.5;
        CFG.det = [];                        % deterministic variant (for now the only one supported)
                                             % (when set to 1, overrides any prob_char setting below)
        CFG.X = {0:5};                       % note assumed to be continuous integer range
        CFG.U = {[-1 1]};                    % note assumed integer values
        CFG.x_energy = 0;
        CFG.x_can = 5;
        CFG.rew_step = 0;
        CFG.rew_energy = 1;
        CFG.rew_can = 5;
        %stochastic properties
        CFG.prob_char = [0, 0];             %chance of [no movement, reverse direction], remainder is normal direction
        
        if nargin >= 2, cfg = varargin{1};
        else cfg = struct; 
        end;
        cfg = parseconfig(cfg, CFG);
        if isempty(cfg.det), cfg.det = all(cfg.prob_char == 0); end;
        % if ~cfg.det, error('CLEANROB_PROBLEM: only deterministic variant supported.'); end;
        
        model = copyfields(cfg, model);
        
        if length(cfg.prob_char) ~= 2
            error('CLEANROB_PROBLEM: one can only give two probability characteristics.');
        elseif sum(cfg.prob_char) >= 1 || sum(cfg.prob_char) < 0
            error('CLEANROB_PROBLEM: wrong values for probability characteristics, sum should be in range [0, 1).');
        end;
        
        % determine terminal states
        model.Xterminal = (model.X{1} == model.x_energy) | (model.X{1} == model.x_can);
        
        % determine minima and maxima
        model.minx = min(model.X{1}); model.maxx = max(model.X{1});
        model.minu = min(model.U{1}); model.maxu = max(model.U{1});
        
        out = model;
        
    % reset to initial state
    case 'reset',
        model = varargin{1};
        reset = varargin{2};
        if isempty(reset), reset = 'rand'; end;
        if isnumeric(reset),
            if size(reset, 2) > 1,      x0 = reset(:, unidrnd(size(reset, 2)));
            else                        x0 = reset;
            end;
        elseif strcmp(reset, 'rand'),
            % reset to random NONTERMINAL state
            X0 = model.X{1}(~model.Xterminal);
            x0 = X0(:, unidrnd(length(X0)));
        end;
        
        out = x0;

end;        % mode SWITCH

% END sample_problem() RETURNING out ===================================================