function ddi_bruteforceopt(cfg)
% Brute-force optimal control computation for DDI problem,
% for a given set of initial states

if nargin < 1, cfg = struct; end;

CFG.problem = 'ddi_problem';
CFG.gamma = 0.95;
CFG.datafile = 'dd_bruteforce';
CFG.steps = 10;
CFG.progdisp = .05;
% CFG.itersave = 10;
CFG.X0 = 'fine';
CFG.model_params = {'QUAD2'};
CFG.verb = 5;
CFG.silent = 0;
CFG.usesubseq = 0;
CFG.improvedatafile = [];

if ischar(cfg), cfg = str2cfg(cfg, fieldnames(CFG)); end;
cfg = checkparams(cfg, CFG);
if cfg.silent, cfg.verb = -Inf; end;
cfg.improve = ~isempty(cfg.improvedatafile);
cfg

% create the model
model = feval(cfg.problem, 'model', cfg.model_params{:});
% get the set of initial states
if ischar(cfg.X0),
    cecfg = feval(cfg.problem, 'ce', cfg.X0);
    X0 = cecfg.X0;
else  % don't care about X0, just get U
    cecfg = feval(cfg.problem, 'ce');
    X0 = cfg.X0;
end;
U = cecfg.U;
gamma = cecfg.gamma;

M = size(U, 2);     % number of actions

% open-loop policies (action sequences) to try for each state
P = dec2base(0:(M^cfg.steps-1), M) - '0' + 1;
NP = size(P, 1);

% precompute discounting vector
disc = cumprod([1 gamma+zeros(1, cfg.steps)]);

% init
N0 = size(X0, 2);
JSTAR = zeros(N0, 1); KSTAR = zeros(N0, 1); PSTAR = nan(N0, cfg.steps);
USTAR = nan(N0, model.q, cfg.steps);
XSTAR = nan(N0, model.p, cfg.steps);
TERM = false(N0, 1);
found = false(N0, 1);     % logical vector storing "optimal found" flags

if cfg.improve,     % improve on existing results by lengthening sequence
    dispx('Loading earlier data', cfg.verb, 1);
    vars = {'JSTAR', 'KSTAR', 'PSTAR', 'USTAR', 'XSTAR', 'TERM', 'found', 'cfg', 'trun'};
    % this will give a warning if TERM was not set in the old datafile
    loadwithprefix(cfg.improvedatafile, 'old', vars);
        
    % these will give errors unless dimensions are correct; that is intended
    % to be so
    oldsteps = oldcfg.steps;
    JSTAR(:) = oldJSTAR;
    KSTAR(:) = oldKSTAR;
    PSTAR(:, 1:oldsteps) = oldPSTAR;        % the rest remain NaN
    USTAR(:, :, 1:oldsteps) = oldUSTAR;     % the rest remain NaN
    XSTAR(:, :, 1:oldsteps+1) = oldXSTAR;   % the rest remain NaN
    found = oldfound;                       % this should be all true if run was completed
    trun = oldtrun;
    if exist('oldTERM', 'var'),
        TERM(:) = oldTERM;
    else        % support for datafiles without TERM flags
        % Safe choice: trajectories for which the length was equal to the max
        % # of steps are assumed not terminated
        TERM(KSTAR == oldsteps) = false;
    end;
    % this will force recomputation for un-terminated states
    found(~TERM) = false;
    % clear loaded variables
    for i = 1:length(vars), clear(['old' vars{i}]); end;
    dispx(['Improving for ' num2str(sum(~found)) ' states'], cfg.verb, 1);
else 
    trun = 0;
end;

dispx('Starting optimization', cfg.verb, 1);
tmark = cputime;

prog = cfg.progdisp;
hits = 0;
for i0 = 1:N0,
    if found(i0), continue; end;
    Jx = zeros(NP, 1); Kx = Jx; term = Jx;
    X = zeros(model.p, cfg.steps+1, NP);
    R = zeros(cfg.steps, NP);
    % loop over possible policies
    for ip = 1:NP,
        X(:, 1, ip) = X0(:, i0);
        for k = 1:cfg.steps,
            [X(:, k+1, ip) R(k, ip) terminal] = feval(model.fun, model, X(:, k, ip), U(:, P(ip, k)));
            if terminal, break; end;    % no other rewards contribute (equivalently, they are all 0)
        end;        % steps loop
        Jx(ip) = disc(1:k) * R(1:k, ip);
        Kx(ip) = k;
        term(ip) = terminal;
    end;            % policies loop
    % find out the best policy & the corresp number of steps, and return
    [jxstar, ipstar] = max(Jx);
    kstar = Kx(ipstar);                                 % length of trajectory
    found(i0) = true;
    JSTAR(i0) = jxstar;                                 % optimal return
    KSTAR(i0) = kstar;
    PSTAR(i0, 1:kstar) = P(ipstar, 1:kstar);            % policy / action indices
    USTAR(i0, :, 1:kstar) = U(:, PSTAR(i0, 1:kstar));   % policy / actual actions
    XSTAR(i0, :, 1:kstar+1) = X(:, 1:kstar+1);          % state traj
    TERM(i0) = term(ipstar);                           % whether best trajectory terminated
    % check if optimal episode terminated
    if ~term(ipstar),
        warning(['Best trajectory did not terminate for x0=[' num2str(X0(:, i0)') ']. Results are unreliable.']); 
%         yn = input('Continue? Y/[N] ');
%         if ~strcmp(upper(yn), 'Y'),
%             disp('Interrupted');
%             return;
%         end;
    end;
	% use Bellman's optimality principle to find optimal subsequences (if
	% configured to do so)
    if cfg.usesubseq,
        Xstar = X(:, 1:kstar+1, ipstar);
        Rstar = R(1:kstar, ipstar);
        remaining = find(~found);
        for k = 2:kstar,
            eq = repmat(Xstar(:, k), 1, length(remaining)) == X0(:, remaining);
            ix = find(all(eq), 1);  % X0 contains unique elements, search for 1 only
            if ~isempty(ix),
                % found a match; determine its original index in X0
                f0 = remaining(ix);
                % mark as found
                found(f0) = true;
                % length of optimal sequence
                fkstar = kstar - k + 1;
                KSTAR(f0) = fkstar;
                % the rhs-s below is displaced k steps to the right
                JSTAR(f0) = disc(1:fkstar) * Rstar(k:kstar);
                PSTAR(f0, 1:fkstar) = P(ipstar, k:kstar);            % policy / action indices
                XSTAR(f0, :, 1:fkstar+1) = Xstar(:, k:kstar+1);      % policy / actual actions
                % use the just-set PSTAR to retrieve the open-loop actions
                USTAR(f0, :, 1:fkstar) = U(:, PSTAR(f0, 1:fkstar));   % policy / actual actions
                % all sub-trajectories terminated IFF main trajectory terminated
                TERM(f0) = TERM(i0);        
                % update the remaining indices
                remaining = find(~found);
                hits = hits+1;
            end;
        end;
    end;
    if i0 / N0 > prog,
        dispx([num2str(prog * 100) '% done, ' num2str(hits) ' optimal subsequences found.'], cfg.verb, 2);
        prog = prog + cfg.progdisp;
        dispx(['Saving intermediate data to ' cfg.datafile], cfg.verb, 1);
        save(cfg.datafile);
    end;
end;                % states loop

trun = trun + cputime - tmark;

dispx(['Optimization complete. Saving to ' cfg.datafile], cfg.verb, 1);
save(cfg.datafile);
    
end

% to compute the optimal policy with nonsaturating terminal states, use:
% cfg = struct; cfg.X0 = flat({-1:2/100:1, -.5:1/100:.5}); cfg.datafile='bfopt_nonsat_optimalh'; cfg.model_params = {'QUAD2', 1, 'nonsat'}; cfg.steps = 10; ddi_bruteforceopt(cfg);
% (need to increase up to 17 to get all optimal trajectories; I used the sequence 13-15-17)

% to compute for p01s02 set of initial states, use:
% cfg = struct; cfg.X0 = 'p01s02'; cfg.datafile='bfopt_nosat_p01s02'; cfg.model_params = {'QUAD2', 1, 'nonsat'}; cfg.steps = 10; ddi_bruteforceopt(cfg);
