%% Optimistic planning experiments for the inverted pendulum
function op_enhancements_hc(mode, expids, custombcfg, customacfg, customplanparams)

%% General config
% clear all; close all;

% Stochastic: batch experiment
BCFGSTOCH = struct;
BCFGSTOCH.alg = 'genmpc';
BCFGSTOCH.verb = 5;
BCFGSTOCH.in = 'planparamsn';
BCFGSTOCH.idfields = {'alg', 'problem', 'planalg', 'X', 'planparamsn', 'rep'};
BCFGSTOCH.startat = 1; BCFGSTOCH.repeat = 6;
BCFGSTOCH.plotindex = 0; BCFGSTOCH.addtocrt = [];
BCFGSTOCH.skipcache = 1;         % we don't need to cache the results, they are easy to compute
%
% algorithm
ACFGSTOCH = struct;
ACFGSTOCH.problem = 'hillcar_problem';
ACFGSTOCH.model_params = {'rewtype=prog progrew=0.01 Ts=0.2 stoch_prob=[.5,.5] stoch_gain=[1.2,.8]'};
ACFGSTOCH.online = 1;
ACFGSTOCH.gamma = 0.95;
ACFGSTOCH.planalg = 'opss';
ACFGSTOCH.planparams.L = 2;
ACFGSTOCH.planparams.U = [-4 4];
ACFGSTOCH.trials = 1;
ACFGSTOCH.maxsteps = 40;
ACFGSTOCH.reset.type = 'seq';
ACFGSTOCH.verb = 3;
ACFGSTOCH.datadir = 'd:\work\m\approxrl\systems\hillcar\op_enhancements_hc';
ACFGSTOCH.X = 'zero';

% Deterministic (as modifications to configs above): batch
BCFGDET = BCFGSTOCH;
BCFGDET.repeat = 1;
% algorithm
ACFGDET = ACFGSTOCH;
ACFGDET.model_params = {'rewtype=prog progrew=0.01 Ts=0.2'};
ACFGDET.planalg = 'opd';
ACFGDET.planparams.L = [];  % meaningless for det case

% default custom configs (empty)
if nargin < 3, custombcfg = struct; end;
if nargin < 4, customacfg = struct; end;
if nargin < 5, customplanparams = struct; end;

    % nested utility function for executing experiment
    function execbatchexp(expid)
        if any(strcmp(expid, expids)) || any(strcmp(expids, {'all', 'any'})),  % if experiment is requested
            bcfg = parseconfig(custombcfg, bcfg); 
            acfg = parseconfig(customacfg, acfg);
            acfg.planparams = parseconfig(customplanparams, acfg.planparams);
            cd(acfg.datadir); 
            % check if we need to call 'process' or 'eval'
            if any(strcmp(mode, 'run')), 
                processbatch(bcfg, acfg); 
            elseif any(strcmp(mode, {'regret', 'exectime', 'plantime', 'depth', 'return'})), 
                bcfg.out = mode;
                evalbatch(bcfg, acfg);
            end;
        end;
    end

%% STOCH: Baseline: no data reuse

bcfg = BCFGSTOCH;
bcfg.mode = 'run';
bcfg.idinfix = 'noreuse';
bcfg.legendprefix = 'No data reuse';

acfg = ACFGSTOCH;
acfg.planparams.n = [500 1000 2000 4000];

execbatchexp(bcfg.idinfix);


%% STOCH: Tree reuse

bcfg = BCFGSTOCH;
bcfg.mode = 'run';
bcfg.idinfix = 'reusetree';
bcfg.legendprefix = 'Tree reuse';

acfg = ACFGSTOCH;
acfg.planparams.n = [500 1000 2000 4000];
acfg.planparams.reusetree = 1;
acfg.planparams.reusetree_maxsize = 0;      % no limit on size

execbatchexp(bcfg.idinfix);
% 
%% STOCH: Memory: study effect of Lipschitz constant

bcfg = BCFGSTOCH;
bcfg.mode = 'run';
bcfg.idinfix = 'lip';
bcfg.legendprefix = 'Memory use';
bcfg.idfields = {BCFGSTOCH.idfields{:}, 'planparamslipschitz'};

acfg = ACFGSTOCH;
acfg.planparams.n = [500 1000 2000 4000];
acfg.planparams.usemem = 1;
acfg.planparams.mem_maxsize = 0;     % unlimited
acfg.planparams.lipschitz = [1 5 10];
% also for 0.1, 3 exper available

execbatchexp(bcfg.idinfix);

%% STOCH: Memory: influence of memory size

bcfg = BCFGSTOCH;
bcfg.mode = 'run';
bcfg.idinfix = 'memsize';
bcfg.legendprefix = 'Memory use, varying size';
bcfg.idfields = {BCFGSTOCH.idfields{:}, 'planparamslipschitz', 'planparamsmem_maxsize'};
bcfg.in = 'planparamsmem_maxsize';

acfg = ACFGSTOCH;
acfg.planparams.n = [500 1000 2000 4000];
acfg.planparams.usemem = 1;
acfg.planparams.lipschitz = 5;
acfg.planparams.mem_maxsize = [0 0.25 0.5 1];
% NOTE! "memmaxsize0" experiments are just copied from the Lipschitz constant study above to avoid
% re-running

execbatchexp(bcfg.idinfix);

%% DET: Baseline: no data reuse

bcfg = BCFGDET;
bcfg.mode = 'run';
bcfg.idinfix = 'detnoreuse';
bcfg.legendprefix = 'No data reuse';

acfg = ACFGDET;
acfg.planparams.n = [20:10:100];

execbatchexp(bcfg.idinfix);

%% DET: Tree reuse

bcfg = BCFGDET;
bcfg.mode = 'run';
bcfg.idinfix = 'detreusetree';
bcfg.legendprefix = 'No data reuse';

acfg = ACFGDET;
acfg.planparams.n = [20:10:100];
acfg.planparams.reusetree = 1;
acfg.planparams.reusetree_maxsize = 0;  % no limit

execbatchexp(bcfg.idinfix);

%% DET: Memory: study of effect of Lipschitz constant when unlimited size

bcfg = BCFGDET;
bcfg.mode = 'run';
bcfg.idinfix = 'detusemem';
bcfg.legendprefix = 'Memory use';
bcfg.idfields = {BCFGDET.idfields{:}, 'planparamslipschitz', 'planparamsmem_maxsize'};
bcfg.in = 'planparamslipschitz';

acfg = ACFGDET;
acfg.planparams.n = [20:10:100];
acfg.planparams.usemem = 1;
acfg.planparams.mem_maxsize = 0;            % no limit
acfg.planparams.lipschitz = [0.1 1:5 10];

% note we use the 'usemem' infix so that experiments are reused between this and the set below, but
% the actual ID is different
execbatchexp('detlip'); 


end
