function out = invpend_problem(what, varargin)
% Integrator problem setup.
%   OUT = DOUBLEINT_PROBLEM(WHAT)
% This function conforms to the specifications established by SAMPLE_PROBLEM.

maxx = [pi; 5*pi]; 
% maxu except when overridden... not all options below take this into account!
maxu = 0.8;     % (not enough to swing upright in one go)
% maxu = 1;     % sufficiently actuated  / no swingup necessary
x0 = [0; 0];

switch what
    case 'info',    % basic info about the model (without creating the actual model)
        info.id = 'invpend'; info.problem = @invpend_problem;
        info.det = 1; info.p = 2; info.q = 1;
        out = info;
        
    case 'model'
        m.id = 'invpend'; m.problem = @invpend_problem;
        m.maxx = maxx;
        m.p = 2;
        m.q = 1;
        m.fun = @invpend_mdp;

        % model type  (note comes after reward argument)
        if nargin < 3 || isempty(varargin{2}),  par = struct;
        else                                    par = varargin{2};
            if any(strcmp(par, {'rot', 'cart'})), % specified just as type (backward compatibility)
                par = struct('type', par);
            end;
        end;
        % init type from supplied config (default is ROTational / normal invpend)
        PAR.type = 'rot'; par1 = parseconfig(par, PAR);
        switch par1.type,
            case 'rot',
                % model parameter defaults: as in Jelmer van Ast's thesis (swingup problem)
                PAR.maxu = maxu;         
                PAR.J = 0.005;
                PAR.K = 0.1;
                PAR.b = 0.01;
                PAR.m = 0.1;
                PAR.l = 0.1;
                
                % init model params
                par = parseconfig(par, PAR);
                m = copyfields(par, m, fieldnames(par));
                m.g = 9.81; % this one is obviously fixed
                % helpers
                m.mgR = m.m * m.g * m.l;    % for compatibility with older models

                % Default sampling time (found by analyzing the sys linearization 
                % in the 2 eq. points)
                Ts = 0.02;
                m.trans = @invpend_trans;
            case 'cart',            
                % model parameters as in Lagoudakis and Parr
                PAR.maxu = 50;
                m.m = 2;
                m.M = 8;
                m.l = 0.5;
                
                % init model params
                par = parseconfig(par, PAR);
                m = copyfields(par, m);
                m.g = 9.81; % this one is obviously fixed
                % helpers
                m.alpha = 1 / (m.m + m.M);
                m.alphaml = m.alpha * m.m * m.l;
    
                % Default sampling time (probably too large, see above)
                Ts = 0.1;
                m.trans = @invpend_cart_trans;
        end;
        
        % reward specification (SECOND argument)
        REW.rewtype = 'lqr';
        REW.Q = diag([5, 0.1]); REW.R = 1;                  % for QR
        REW.zeroreward = 1; REW.boxrew = [5*pi/180; 0.1];   % for BOX
        REW.sigma = diag([m.maxx; m.maxu]/3);               % for GAUSS
        if nargin < 2 || isempty(varargin{1}),  rew = REW;  % use defaults
        else
            rew = varargin{1};
            % support reward specified simply by type
            if ischar(rew) && any(strcmp(rew, {'lqr', 'box', 'gauss', 'fail'})),
                rtype = rew; rew = REW; rew.rewtype = rtype; % the rest remain at defaults
            else    % full-fledged configuration parsing
                rew = parseconfig(varargin{1}, REW);
            end;
        end;
        
        switch rew.rewtype,
            case 'lqr',
                m = copyfields(rew, m, {'rewtype', 'Q', 'R'});
                m.maxr = m.maxx' * m.Q * m.maxx + m.R * m.maxu * m.maxu;
            case 'gauss',
                m = copyfields(rew, m, {'rewtype', 'sigma'});
                m.W = inv(2 * m.sigma * m.sigma);
                m.maxr = abs(-1 + exp(-[m.maxx'; m.maxu]' * m.W * [m.maxx'; m.maxu]));
                % to run a fuzzyQI exper
                % clear fc; fc.run = 1; fc.model_params={'gauss'}; fc.datafile='fzip_fine_gauss'; fc.problem='invpend_problem'; fuzzyqi(fc);
            case 'box',             
                m = copyfields(rew, m, {'rewtype', 'boxrew', 'zeroreward'});
                m.maxr = m.zeroreward;
                % swingup:
                % clear fc; fc.run = 1; fc.model_params={'box'}; fc.datafile='fzip_fine_box'; fc.problem='invpend_problem'; fuzzyqi(fc);
                % stabilization (no swing necessary):
                % clear fc; fc.run = 1; fc.model_params={'box','maxu=2'}; fc.datafile='fzip_fine_box_u2'; fc.problem='invpend_problem'; fuzzyqi(fc);
            case 'fail',            % failure-type reward, 0 everywhere except terminal state
                m = copyfields(rew, m, {'rewtype'});
                m.maxx(1) = pi/2;   % IMPORTANT! angle cannot be larger than pi/2
                m.maxr = 1;
        end;
        
        % default integration parameters (method chosen by letting invpend fall and using
        % odetest ode45 and ode4N1 are nearly identical, so ode4N1 suffices)
        ODE.Ts = Ts;                    % scalar, sampling time      
        ODE.odemethod = 'fixedode';     % 'varode', 'fixedode', 'eul'
        ODE.odesolver = @ode4;          % solver function
        ODE.odesteps = 1;               % # of steps per sample time (ignored for varode)
        if nargin < 4 || isempty(varargin{3}),
            ode = ODE;
        else
            % (possibly / some) user-specified integration parameters
            ode = str2cfg(varargin{3}, ODE); ode = checkparams(ode, ODE);
        end;
        m = copyfields(ode, m);
        % method-specific settings
        switch(ode.odemethod)
            case 'varode',
                m.odet = [0 m.Ts/2 m.Ts];
                m.odeopt = odeset;
            case 'fixedode',
                m.odet = 0 : m.Ts / m.odesteps : m.Ts;
            case 'eul',   
                % no settings
        end;

        out = m;
        
    case 'linmod',
        % linearized cont-time model around the unstable equilibrium; 
        % gives only cart-less linearized model
        if nargin < 2 || isempty(varargin{1}), m = invpend_problem('model');
        else m = varargin{1};
        end;
        % which equilibrium point [x; u]
        if nargin < 2, fp = [0 0 0]';
        else fp = varargin{2};
        end;
        
        ssm = ss([0 1; m.mgR/m.J*cos(fp(1)) -m.b/m.J], [0; m.K/m.J], [1 0], 0);
        tfm = tf(ssm);
        step(tfm);
        out = tfm;
    
    case 'fuzzy'
%         cfg.gamma = gamma;
        
        % process arguments for fuzzy grids
        if nargin < 2, xg = 'fine';         % Default is fine scale
        else xg = varargin{1}; end;
        if nargin < 3, Npoints = 61;        % Default # log points on each axis (including 0)
        else Npoints = varargin{2}; end;
        
        % WARNING! global maxu is used here, even though the model maxu might be different!!!
        switch xg,
            case 'fine',        % use fine state grids
                cfg.xgrids = {-pi:2*pi/180:pi, -maxx(2):pi/8:maxx(2)};
                cfg.ugrids = {symloggrid(15, maxu, -1)};
            case 'log',
                cfg.xgrids = {symloggrid(Npoints, maxx(1)), symloggrid(Npoints, maxx(2))};
                cfg.ugrids = {symloggrid(15, maxu, -1)};
            case 'coarse',
                cfg.xgrids = {-pi:30*pi/180:pi, -maxx(2):pi/2:maxx(2)};
                cfg.ugrids = {symloggrid(5, maxu, -1)};
        end;
        cfg.x0 = x0;
    
        out = cfg;

    case 'X0',
        if nargin < 2, xt = 'coarse';
        else xt = varargin{1}; 
        end;
        switch xt,
            case 'coarse',
                out = {-pi:30*pi/180:pi, -5*pi:pi:5*pi};
            otherwise,
                error(['Unknown X0 type [' xt ']']);
        end;
        
    case {'lspi', 'lspe'},
%         cfg.gamma = gamma;
        cfg.x0 = x0;
        out = cfg;

end;


% END doubleint_problem(), RETURNING out ====================================