function uquant_study(cfg)
% Assumes 2-state, 1-output, system as implemented currently

% default arguments
if nargin < 1, cfg = ''; end;
% model setup
CFG.control = 'direct';
CFG.type = 'ls';
% quantization settings
CFG.dx = .2;            % state quantization step
% CFG.ures = 1:10;
CFG.ures = 1:8;         % number of quanta - 1 in an equidistant control quantization
% RL (QI) settings
CFG.gamma = 0.98;
CFG.eps = 1e-4;
CFG.maxiter = 1000;
% result eval settings
CFG.interph = 0;        % if interpolated policy
CFG.replay = 0;         % replay from one initial state
CFG.return = 0;         % plot return map (computationally intensive)
CFG.dx0 = .25;          % step for policy evaluation
CFG.mc_maxsteps = 0.01; % < 1: precision evaluation in MC return; >= 1: number of steps
% misc
CFG.cachefile = 'uqcache';
CFG.ignorecache = 0;

% process config
if ischar(cfg), cfg = str2cfg(cfg, fieldnames(CFG)); end;
cfg = checkparams(cfg, CFG);
% fcfg.grayscale = grayscalefromconfig(cfg);

% a dummy model used in settings below
m = linsys_problem('model', cfg.type, cfg.control);

% derived settings in config
if cfg.mc_maxsteps < 1,         % max steps is relative return error, use it to compute maxsteps
    cfg.mc_maxsteps = ceil(log(cfg.mc_maxsteps * (1-cfg.gamma) / m.maxr) / log(cfg.gamma));
end;


% grid settings
dx = cfg.dx;
xgrids = {-m.maxx(1):dx:m.maxx(1), -m.maxx(2):dx:m.maxx(2)};
switch cfg.control,
    case 'direct',
        usteps = m.maxu ./ cfg.ures;
    case 'lsf',
        usteps = [m.maxu(1) ./ cfg.ures; m.maxu(2) ./ cfg.ures];
end;

Nexp = length(usteps);

% prepare return computation data
if cfg.return,
    if isscalar(cfg.dx0), 
        cfg.dx0 = cfg.dx0 + zeros(m.p, 1);
    end;
    X0 = cell(m.p, 1); N0 = zeros(1, m.p);
    for p = 1:m.p,
        X0{p} = -m.maxx(p) : cfg.dx0(p) : m.maxx(p);
        N0(p) = length(X0{p});
    end;
    J = zeros([Nexp prod(N0)]);
end;

% run control resolutions loop
for i = 1:Nexp,
    
	du = usteps(:, i);
    datafile = makefilename(m, dx, du, cfg);
    
    % --- RUN -----------------------
    % run if needed
    if ~exist([datafile '.mat'], 'file'),        
        fcfg = struct;
        fcfg.problem = 'linsys_problem';
        fcfg.model_params = {cfg.type, cfg.control};
        fcfg.run = 1;
        fcfg.maxiter = cfg.maxiter;
        fcfg.eps = cfg.eps;
        fcfg.gamma = cfg.gamma;
        fcfg.datafile = datafile;
        fcfg.xgrids = xgrids;
        switch cfg.control,
            case 'direct',
                fcfg.ugrids = {-m.maxu:du:m.maxu};
            case 'lsf',
                fcfg.ugrids = {-m.maxu(1):du(1):m.maxu(1), -m.maxu(2):du(2):m.maxu(2)};
        end;

        fuzzyqi(fcfg);
    end;
    
    % --- REPLAY -----------------------
    if cfg.replay,
        figure;
        fcfg = struct;
        fcfg.replay = 1;
        fcfg.problem = 'linsys_problem';
        fcfg.x0 = [-0.5 0]';
        fcfg.datafile = datafile;
        fcfg.silent = 1;
        fcfg.interph = cfg.interph;
        fcfg.tend = 5;
        rhist = fuzzyqi(fcfg);

        plothistory(rhist);
%         pause;
    end;
    
    % --- RETURN plot -----------------------
    if cfg.return,

        % load interesting data from file
        load(datafile, 'X', 'U', 'theta', 'model');
        
        % make cache field, verify if cache exist and is valid
        cfield = makealphanum(datafile);
        if cfg.interph,     % add interpolation switch
            cfield = [cfield 'interph'];
        end;
        if exist([cfg.cachefile '.mat'], 'file'), 
            load(cfg.cachefile, 'CACHE');
            if isfield(CACHE, cfield) && ~cfg.ignorecache,
                c = CACHE.(cfield);
                % cache validity: same model, same set of initial states
                cachevalid = all(size(flat(c.X0)) == size(flat(X0))) ...
                    && all(all(flat(c.X0) == flat(X0))) ...
                    && isequal(model, c.model);
            else cachevalid = 0;    % because there's nothing saved for this datafile / experiment
            end;
        else cachevalid = 0;        % because there's no cache file
        end;
        if cachevalid,  % cache valid, just get scores
            J(i, :) = c.J(:);
        else            % cache invalid; need to run simulations
            % run trajectories from each initial state in X0
            [eJ, J(i, :)] = mc2_fuzzyq(X, flat(U), theta, model, flat(X0), cfg);
            c = struct;
            c.model = model; c.X0 = X0; c.J = J(i, :);
            CACHE.(cfield) = c;
            save(cfg.cachefile, 'CACHE');
        end;
        
        % plot result
        fh = figure;
        set(fh, 'NumberTitle', 'off', 'Name', datafile);
        ph = pcolor(X0{1}, X0{2}, reshape(J(i, :), N0)');
        set(ph, 'LineStyle', 'none'); hold on; 
        % axis between 0 and maximum theoretical reward
        caxis([0 m.maxr/(1-cfg.gamma)]);
        colormap(gray);
        
%         if i > 2,
%             fh = figure;
%             set(fh, 'NumberTitle', 'off', 'Name', ['x0 w/ perf decrease i=' num2str(i)]);
%             Js = (diff(J(i-1:i, :), 1) < 0) + 0;
%             ph = pcolor(X0{1}, X0{2}, reshape(Js, N0)');
%             set(ph, 'LineStyle', 'none'); hold on; 
%             colormap(gray);
%         end;
    end;

    
end;

% fh = figure(20); clf;
% set(fh, 'NumberTitle', 'off', 'Name', 'x0 w/ perf decrease');
% Js = any(diff(J, 1) < 0) + 0;
% ph = pcolor(X0{1}, X0{2}, reshape(Js, N0)');
% set(ph, 'LineStyle', 'none'); hold on; 
% colormap(gray);
    
end         % uquant_study() -------------------------------------------------

function datafile = makefilename(m, dx, du, cfg)
if strcmp(cfg.control, 'lsf'),
    datafile = ['ls_Ts' num2str(m.Ts) '_umax' num2str(m.maxu(1)) 'X' num2str(m.maxu(2)) ...
        '_dx' num2str(floor(dx*1000)/1000) ...
        '_du' num2str(floor(du(1)*1000)/1000) 'X' num2str(floor(du(1)*1000)/1000)];
else
    datafile = ['ls_Ts' num2str(m.Ts) '_umax' num2str(m.maxu) '_dx' num2str(floor(dx*1000)/1000) '_du' num2str(floor(du*1000)/1000)];
end;
datafile(datafile == '.') = '_';

end         % makefilename() --------------------------------------------------

