function [xplus, rplus, terminal] = nav_mdp(m, x, u)
% Implements the discrete-time dynamics of the navigation problem.
%  [XPLUS, RPLUS, TERMINAL] = NAV_MDP(M, X, U)
%
% This function conforms to the specifications established by SAMPLE_MDP.

% limit torque
u = max(-m.phys.maxu, min(m.phys.maxu, u));

if m.disc.method == 1,          % ODE
    odet = m.disc.odet;           % we only care about solution at final time
	[odet odey] = feval(m.disc.odesolver, @nav_trans, odet, x, m.disc.odeopt, m.phys, u);
    xplus = odey(end, :)';      % pick up the state at t + Ts

elseif m.disc.method == 2,      % discrete ODE
    odet = m.disc.odet;
    odey = feval(m.disc.odesolver, @nav_trans, odet, x, m.phys, u);
    xplus = odey(end, :)';      % pick up the state at t + Ts
        
elseif m.disc.method == 3,      % Euler, x(k+1) = x(k) + Ts f(x(k), u(k))
    xplus = x + m.disc.Ts * nav_trans(0, x, m.phys, u);
end;

% Bound the state: position between 0 and size; speed +-max
xplus = max(-m.phys.maxx, min(m.phys.maxx, xplus));

% Reward - QR component, plus optionally the band reward
if all(abs(x) <= m.goal.zeroband),
    rplus = m.goal.zeroreward;
else 
    rplus = -x' * m.goal.Q * x - u' * m.goal.R * u;
end;

% old (not directly comparable to binary) method of computing reward
% rplus = -x' * m.goal.Q * x - u' * m.goal.R * u + all(abs(x) <= m.goal.zeroband) * m.goal.zeroreward;
terminal = 0;       % task is continuing
% END FUNCTION -----


function xdot = nav_trans(t, x, p, u)
%  Implements the transition function of the double integrator
%
%  Parameters:
%   T           - the current time. Ignored, added for compatibility with
%               the ode function style.
%   X           - the current state
%   P           - the physical parameters
%   U           - the command
%
%  Returns:
%   XDOT        - the derivative of the state
%

b = feval(p.damping.fun, p.damping, x);
xdot = [x([3;4]); -b*x([3;4]) + u];
