function [xplus, rplus, terminal, pplus, refxplus] = rarm_mdpstandalone(m, x, u, w, refx)
% Discrete-time dynamics of the robot arm
%  [XPLUS, RPLUS, TERMINAL] = RARM_MDP(M, X, U)
%  [XPLUS, RPLUS, TERMINAL, PPLUS] = RARM_MDP(M, X, U)
%   or
%  [XPLUS, RPLUS, TERMINAL, [], REFXPLUS] = RARM_MDP(M, X, U, [], REFX)
% Standalone, new version (does not require the m/rarm directory).
% Can be wrapped by stochact_mdpwrapper to generate stochastic-action based transitions.
% Returns a probability of "1" as PPUS, if this argument is requested.
% Last form should be used if a reference model is being used
% This function conforms to the specifications established by SAMPLE_MDP.

% limit torque
u = max(-m.maxu, min(m.maxu, u));

% integrate numerically along the sample time
if m.odemethod(1) == 'o',         % ODE
	[odet odey] = feval(m.odesolver, @rarm_transstandalone, m.odet, x, m.odeopt, m, u);
    xplus = odey(end, :)';      % pick up the state at t + Ts

elseif m.odemethod(1) == 'f',     % fixed-step ODE
    odey = feval(m.odesolver, @rarm_transstandalone, m.odet, x, m, u);
    xplus = odey(end, :)';      % pick up the state at t + Ts
        
elseif m.odemethod(1) == 'e',     % Euler, possibly multistep
    dt = m.Ts / m.odesteps; xplus = x;
    for i = 1:m.odesteps,
        xplus = xplus + dt * rarm_transstandalone(0, x, m, u);
    end;
end;

% Normalized state
if m.wrap,  % wrap angles in [-pi, pi), bound velocities
    xplus([1 3]) = mod(xplus([1 3]) + pi, 2*pi) - pi;
    xplus([2 4]) = max(-m.maxomega, min(m.maxomega, xplus([2 4])));
else        % no wrapping, just bound everything
    xplus = max(-m.maxx, min(m.maxx, xplus));
end;

% Compute reward
if m.rewtype(1) == 'l',             % LQR
    rplus = - x' * m.Q * x  - u' * m.R * u;
elseif m.rewtype(1) == 'n',             % normalized LQR
    rplus = 1 - (x' * m.Q * x  + u' * m.R * u) / m.nonnorm_maxr;
elseif m.rewtype(1) == 'v',         % vf/vectorfield, follow vector field in cartesian coords
    % Note that this reward function is for a horizontally oriented arm:
    % angle [0,0] corresponds to arm maximally extended to the right in the X,Y plane
    % positions in cartesian coordinates
    pos = m.tocartesian(m, x); posplus = m.tocartesian(m, xplus);
    % reference position according to vector field
    [odet odey] = ode45(m.vectorfield, [0 m.Ts], pos, [], m.vectorfield_params{:});
    refpos = odey(end, :)';
    % distance gives the penalty
    rplus = -norm(posplus - refpos)^2;
elseif m.rewtype(1) == 'r',         % reference model for end-effector in cartesian coords 
    [refpos, refxplus] = m.refmodel(m, refx, m.refmodel_params{:});
    pos = m.tocartesian(m, x);
    rplus = 1 - sum((pos - refpos).^2) ./ m.nonnorm_maxr;
end;

terminal = 0;       % task is continuing

% support "distribution" mode
if nargout >= 4, pplus = 1; end;

end         % rarm_singletransition


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Implements the transition function of the two-link robot arm.
function xdot = rarm_transstandalone(t, x, m, u)
%  Implements the transition function of the two-link robot arm.
%  Gravity is neglected if m.neglectG is set
%
%  Parameters:
%   T           - the current time. Ignored, added for compatibility with
%               the ode function style.
%   X           - the current state
%   M           - the model structure for the robot arm
%   U           - the vector of joint torques
%
%  Returns:
%   XDOT        - the derivative of the state
%

s2 = sin(x(3)); c2 = cos(x(3));
M = [m.P1 + m.P2 + 2 * m.P3 * c2,   m.P2 + m.P3 * c2; ...
     m.P2 + m.P3 * c2,              m.P2];
C = [m.b(1) - m.P3 * x(4) * s2,     -m.P3 * (x(2) + x(4)) * s2; ...
     m.P3 * x(2) * s2,              m.b(2)] * x([2 4]);
if m.neglectG,          % no gravity
    G = [0; 0];
else
    G = [-m.g1 * sin(x(1)) - m.g2 * sin(x(1) + x(3)); ...
         -m.g2 * sin(x(1) + x(3))];
end;
if m.type(1) == 'a',    % acrobot: only middle joint actuated, add a zero for the first one
    u = [0; u];
elseif m.type(1) == 'r',% acrobot: only first joint actuated, add a zero for the second
    u = [u; 0];
end;

xdot = x;
xdot([1 3]) = x([2 4]);
xdot([2 4]) = inv(M) * (u - C - G);

end         % rarm_transstandalone

