function out = resonatingarm_problem(what, varargin)
% Integrator problem setup.
%   MODEL = RESONATINGARM_PROBLEM(MODE, [CONFIG])
% This function conforms to the specifications established by SAMPLE_PROBLEM.
% Second argument for 'model' mode is the model config.

switch what
    case 'shortid', 
        out = 'resarm';
    
    case 'model'
        % #### Standard fields
        CFG.id = 'resonatingarm';
        CFG.p = 2; CFG.q = 1;
        CFG.maxx = [2; 2*pi];
        CFG.maxu = 2;
        CFG.det = 1;
        CFG.userefmodel = 0;
        CFG.fun = @resonatingarm_mdp;
        CFG.visualizefun = @dc_ip_visualize;
        CFG.plotfun = @dc_plot;
        
        % #### Physical parameters -- those that are unused are commented out
        % model type
        CFG.type = 'inclined';
        CFG.angle = 0.125*pi; % rad         for 'inclined': angle of the arm plane of rotation
        % friction mechanical specs
        CFG.m1 = 0.6;       % kg            Mass of bar 1
        CFG.m2 = 0.6;       % kg            Mass of bar 2 including 1kg payload
        CFG.Ma = 1;         % kg            Added mass at the end point
        CFG.I1 = 0.03;      % kg m^2        Mass moment of inertia about CoM of bar 1
        CFG.I2 = 0.03;      % kg m^2        Mass moment of inertia about CoM of bar 2
        CFG.lg1 = 0.5;      % %             Percentage of bar 1 equals distance to CoM of the bar
        % CFG.lg2 = 0.5;      % %             Percentage of bar 2 equals distance to CoM of the bar
        CFG.l1 = 0.4;       % m             Length of bar 1
        CFG.l2 = 0.4;       % m             Length of bar 2
        CFG.g = 9.81;       % m/s^2         Gravity constant
        % there is no 2nd DOF
        % CFG.m2 = 0;
        % CFG.I2 = 0;
        CFG.lg1 = 0.9;
        CFG.m1 = CFG.m1 + CFG.Ma;
        % motor specs
        % CFG.kt = 0.0259;        % Nm/A          Motor constant
        % CFG.Ohm  = 0.61;        % ohm           Terminal resistance
        % CFG.L  = 0.119;         % mH            Terminal inductance
        % CFG.eta_out = 0.87;     %               Gearbox efficiency from motor to system
        % CFG.eta_in = 0.87;      %               Gearbox efficiency from system to motor
        CFG.n = 54;             % rad/rad       Gearbox ratio
        % CFG.Tmax =CFG.maxu/(CFG.eta_out*CFG.n);
        CFG.Ir = 33.5E-7;       % kg m^2      
        % friction 
        CFG.cfd1 = 0.0003;      % Nms/rad       dynamic friction coefficient at joint 1
        CFG.cfd2 = 0.0006;      % Nms/rad       dynamic friction coefficient at joint 2
        CFG.cfs1 = 0.4793;      % Nm            static friction at joint 1
        CFG.cfs2 = 0.6;         % Nm            static friction at joint 2
        % spring mechanism
        CFG.r1 = 0.105;         % m         radius of large pulley
        CFG.r2 = 0.02;          % m         radius of small pulley
        CFG.R = 5.4;            % rad/rad   transfer ratio
        CFG.k = 150;            % N/m       stiffness of the spring. ****Set to zero to remove spring mechanism****
        CFG.l0 = 0.1;           % m         initial length of the spring
        CFG.dl = 0.1;           % m         distance between the pulleys
        CFG.d = 0.005;          % m         diameter of the axis at the ends of the spring
        CFG.F0 = 6;             % N         Pre-stress in the spring
        % stochasticity
        CFG.stoch_prob = [];    % if stochastic: probabilities with which...
        CFG.stoch_gain = [];    % the action variable is multiplied with these values
        
        % #### Numerical integration
        % default config leads to errors in the third digit w.r.t. to ode45
        CFG.Ts = 0.05;                 % scalar, sampling time      
        CFG.odemethod = 'fixedode';    % 'varode', 'fixedode', 'eul'
        CFG.odesolver = @ode4;         % solver function
        CFG.odesteps = 10;             % # of steps per sample time (ignored for varode)

        % #### Reward config
        CFG.rewtype = 'nlqrref';       % 'expref'onential or 'nlqrref'
        CFG.xgoal = [0.85; 0];
        CFG.Qrew = diag([1 0]); CFG.Rrew = 0;
        CFG.Crew = diag([2 0.07 0.001]);   % ???     
        
        % ######## Process the config into the model structure
        if nargin == 1 || isempty(varargin{1}), model = CFG;
        else                                    model = parseconfig(varargin{1}, CFG);
        end;
        
        % error checks
        if ~any(strcmp(model.type, {'inclined'})), error('Model type [%s] unsupported', model.type); end;
        if any(model.xgoal < -model.maxx) || any(model.xgoal > model.maxx), error('Goal state outside bounds'); end;

        % physics: process stochastic options
        model.det = isempty(model.stoch_prob);
        if ~model.det, model = stochact_mdpwrapper(model, 'prepare'); end;
        
        % reward: process derived params 
        switch model.rewtype,
            case 'expref',
                model.maxr = 1; % nothing else needed, the reward is already in the interval [0, 1]
            case 'nlqrref',     % LQR normalized to interval [0, 1]
                maxxdiff = max(model.maxx - model.xgoal, model.xgoal - (-model.maxx));
                model.nonnorm_maxr = maxxdiff' * model.Qrew * maxxdiff + model.Rrew * model.maxu^2;
                model.maxr = 1;
        end;
        % num int: process derived params 
        switch(model.odemethod)
            case 'varode',
                model.odet = [0 model.Ts/2 model.Ts];
                model.odeopt = odeset;
            case 'fixedode',
                model.odet = 0 : model.Ts / model.odesteps : model.Ts;
        end;
        
        out=model;
      
    case 'reset',   % handle state reset, only single-value supported
        % model = varargin{1};
        % res = varargin{2};                 
        out = varargin{2};
        
    case {'fuzzy', 'lspi', 'lspi'},     % just for compatibility, don't configure anything
        out = struct;
end;

end % END doubleint_problem(), RETURNING out ====================================


