function [xplus, rplus, terminal] = sw_mdp(model, x, u)
% Implements the dynamics and reward function of the simple walker.
%  [XPLUS, RPLUS, TERMINAL] = SW_MDP(MODEL, X, U)
%
% This function conforms to the specifications established by SAMPLE_MDP.

% 
% limit torque
u = max(-model.maxu, min(model.maxu, u));

% simulate
timespan  = [0 model.Ts];
options  = odeset('events', @sw_events);
[t,x] = ode45(@sw_dynamics, timespan, x, options, model, u);
while t(end) < model.Ts
    % sw_event interrupted ode45 -- there was an impact!
    x_afterimpact = sw_impact(x(end,:));
    % continue to complete the timestep
    timespan = [t(end) model.Ts];
    [t,x] = ode45(@sw_dynamics, timespan, x_afterimpact, options, model, u);
    % this loop could be a while loop to catch multiple impacts within one
    % timestep
end
xplus = x(end,:);

% check if fallen
fall = abs(x(1)) >= pi/2;

% compute reward
rplus = fall * model.rfall + model.ru * abs(u);

% make xplus column vector, and bound it
% xplus = xplus(:);
xplus = max(model.xrange(:, 1), min(model.xrange(:, 2), xplus(:)));

terminal = 0;       % task is continuing

% END FUNCTION sw_mdp RETURNING xplus, rplus ==================================



% %%%%%%%%%%%%%%%%%%%%%%%%
% LOCAL FUNCTIONS

function xnew = sw_impact(xold)
% hopefully I am calculating the impacts as well as swapping the stance and
% swing leg definitions here....
thetaold    = xold(1);
% phiold      = xold(2);
thetadotold = xold(3);
% phidotold   = xold(4);

thetadotnew = cos(2*thetaold)*thetadotold;
phidotnew   = cos(2*thetaold)*(1 - cos(2*thetaold))*thetadotold;
thetanew    = -thetaold;
phinew      = -2*thetaold;

xnew = [thetanew, phinew, thetadotnew, phidotnew];


function xd = sw_dynamics(t, x, m, u)
% Implements the dynamics of the simple walker
theta    = x(1);
phi      = x(2);
thetadot = x(3);
phidot   = x(4);

qdd = [ sin(theta-m.slope);
        sin(phi)*(thetadot^2-cos(theta-m.slope)) + sin(theta-m.slope) + u];
    
xd = [thetadot, phidot, qdd(1), qdd(2)].';


function [value, isterminal, direction] = sw_events(t, x, m, u)
% Swing leg impact detection
theta    = x(1);
phi      = x(2);
value = cos(theta) - cos(phi-theta);
isterminal = (theta<-0.03) & (x(3) < 0) & (x(2) < -0.01);
direction = -1;
