function figh = sw_plot(cfg)
% Plot various information regarding the simple walker problem
%   FIGH = SW_PLOT(CFG)
% Parameters:
%   CFG         - history for a simple trajectory plot;
%               or full configuration of plot, see commented defaults for explanations
%
% Returns:
%   FIGH        - an (array of) handles to the created figure(s)

% default arguments
if nargin < 1, cfg = ''; end;
% support for default plot-function calling mode -- i.e. only history structure
% if isstruct(cfg) && isfield(cfg, 't') && isfield(cfg, 'x'),
%     hist = cfg;
%     cfg = struct();
%     cfg.trajectory = 1;
%     cfg.datasource = 'caller';
%     cfg.hist = hist;
% end;

CFG.datafile   = '';    % data source can be given as a data file here as well
% What to plot
CFG.centers = 1;
CFG.policy = 0;         % if a slice thru the policy should be plotted
CFG.axes = [NaN 0 NaN 0]';     % TODO
% plot configuration
CFG.arrowscale = .2;    % how long should be the command be scaled to obtain arrow length in policy plot
CFG.axisstep = 0.1;     % the grid step for plotting the policy
CFG.markersize = 3;     % the marker size for centers, grid points, trajectory samples etc.
CFG.trajstyle = '-';    % the line style for trajectory plot
CFG.trajcolor = 'k';    % the color for trajectory plot
CFG.showreturn = 1;     % whether to show the return near the trajectory
% save configuration
CFG.plottarget = 'screen';     % 'latex', 'beamer', 'screen', ''
% Save implemented for: policy
CFG.savedir = pwd;
                        % path for saving figure
CFG.savefig = 'sw';    % filename for saving figure

% process config
if ischar(cfg), cfg = str2cfg(cfg, fieldnames(CFG)); end;
cfg = checkparams(cfg, CFG);
cfg.grayscale = grayscalefromconfig(cfg);
% cfg           % feedback on config

% load needed vars from datafile
vars = {'model', 'X', 'U', 'DIMS', 'XMFS', 'theta'};
load(cfg.datafile, vars{:});

% grayscale styles
gs.innerc = .25 * [1 1 1];
gs.centerc = 0 * [1 1 1];
% color styles
cs.innerc = 'b';
cs.centerc = 'r';
% set style
if cfg.grayscale, sty = gs; 
else sty = cs; end;

% labels
extraprop = {'FontSize',13};
labels.x = {'\theta','\phi','\dot\theta','\dot\phi'};

% fig size -- only takes effect for 2D figures 
switch cfg.plottarget
    case 'latex',   
        cfg.figsize = [400 400];
    case 'beamer',  
        cfg.figsize = [400 400];
    otherwise       % default
        cfg.figsize = [];
end; 

% Process plot paths
figh = [];

% active axes
a = find(isnan(cfg.axes));
% fixed axes
f = find(cfg.axes == 0);
% ------ Fuzzy policy
if cfg.policy,
    figh(end+1) = figure; 
    if ~isempty(cfg.figsize),
        set(gcf, 'Position', [0 0 cfg.figsize]);
        movegui(gcf, 'center');
    end;
   
    [Qstar ui] = max(theta, [], 2); clear Qstar;
    ui = lin2ndi(ui, DIMS.dimu);
    % compute optimal policy
    hstar = zeros(DIMS.N, DIMS.q);
    for q = 1:DIMS.q,
        hstar(:, q) = U{q}(ui(:, q));
    end;
    
    % compute the policy on a fine grid
    % ensure the grid points are on that grid
    x1 = sortx([ X{a(1)}  X{a(1)}(1) : cfg.axisstep : X{a(1)}(end) ]);
    x2 = sortx([ X{a(2)}  X{a(2)}(1) : cfg.axisstep : X{a(2)}(end) ]);
    hslice = zeros(length(x1), length(x2));
    for i = 1:length(x1),
        for j = 1:length(x2),
            % set complete state
            x = cfg.axes; x(a(1)) = x1(i); x(a(2)) = x2(j);
            % compute mu
            mu = mdegs(x(1), XMFS{1});
            for p = 2:DIMS.p,
                mu = mu(:) * mdegs(x(p), XMFS{p})';
            end;
            phi = find(mu); mu = mu(phi);
            % compute optimal action
            hslice(i, j) = (mu' * hstar(phi, :))';
        end;
    end;
    % plot the policy
    mesh(x1, x2, hslice'); 

    if cfg.centers,
        % extract slice through policy
        hstar = reshape(hstar, [DIMS.dimx DIMS.q]);
        % form a string index to pick up the slice from hstar
        hind = '';
        for p = 1:DIMS.p,
            if any(p == a),     hind = [hind ':,'];
            else                hind = [hind num2str(find( X{p} == cfg.axes(p) )) ','];
            end;
        end;
        hind = hind(1:end-1);
        hgrid = squeeze(eval([ 'hstar(' hind ')' ]));
        
        % plot the slice through the policy grid
        surface(X{a(1)}, X{a(2)}, hgrid', 'EdgeColor', 'none', 'FaceColor', 'none', ...
            'Marker' ,'o', 'MarkerFaceColor', sty.centerc, 'MarkerSize', cfg.markersize); 
    end;

    % x and y labels
    string2latexlabel(labels.x{a(1)}, extraprop, @xlabel);
    string2latexlabel(labels.x{a(2)}, extraprop, @ylabel);
    
    % z label is more complex, indicates also fixed values for slice
    zl = 'h*(';
    for p = 1:DIMS.p,
        if any(p == a),     zl = [zl labels.x{p} ','];
        else                zl = [zl num2str(cfg.axes(p)) ','];
        end;
    end;
    zl(end) = ')';      % replace last, uneeded comma with closing parenthesis
    % put the label
    string2latexlabel(zl, extraprop, @zlabel);
    
    % finalize image
%     axis equal image
%     set(gcf, 'Units', 'normalized');

end;

% % ------ Fuzzy value function
% if cfg.Q,
%     figh(end+1) = figure; clf;
%     
%     % plot in red points the optimal values in the position grid points
%     % just pick up the indicated speed slice
%     sliceix = [find(X{3} == speedslice(1)) find(X{4} == speedslice(2))];
%     theta = reshape(max(theta, [], 2), DIMS.dimx);
%     % retrieve speed slice
%     theta = squeeze(theta(:, :, sliceix(1), sliceix(2)));
%     mesh(X{1}, X{2}, theta');
%     surface(X{1}, X{2}, theta', 'EdgeColor', 'none', 'Marker' ,'o', 'MarkerFaceColor', 'r', 'MarkerSize', cfg.markersize);
%     xlabel(labels.x{:}); ylabel(labels.y{:}); zlabel(labels.q{:});
%     
%     % title
%     switch cfg.plottarget,
%         case {'latex', 'beamer'},
%         case {'screen', ''},  
%             title(['$Q^*(C_X, C_Y)$ for $[\dot{C}_X, \dot{C}_Y] = [' ...
%                 num2str(speedslice(1)) ',' num2str(speedslice(2)) ']$'], 'Interpreter', 'LaTeX');
%     end;
% end;
% 
% save last figure if indicated
saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);


% END nav_plot RETURNING array of figure handles =================================================
