% Script to test various kNN implementations

%% Initialize
clear all; close all;
Nruns = 1;          % how many times to run tree creation & query (to get average run times)

% small test for visualization
maxx = [pi 10*pi]'; % dimension of (symmetrical) input space
D = length(maxx);   % dimensionality
% S = [1 1];  % distance metric
S = 1 ./ maxx;  % distance metric
Ns = 500;           % number of samples to build the tree from
Nq = 10;             % number of query samples
K = 20;

% % large test for execution time or extensive error checks
% maxx = [pi 10*pi pi 15*pi 10 30]';  % dimension of (symmetrical) input space
% D = length(maxx);   % dimensionality
% S = ones(D, 1); % distance metric
% Ns = 50000;         % number of samples to build the tree from
% Nq = 10000;         % number of query samples
% K = 10;

Xs = rand(D, Ns) .* repmat(maxx, 1, Ns);
Xq = rand(D, Nq) .* repmat(maxx, 1, Nq);


%% Test CPP kd-tree implementation -- http://www.mathworks.in/matlabcentral/fileexchange/21512-kd-tree-for-mat
tcpp = struct;
rcpp = struct; rcpp.name = 'CPP kd-tree'; rcpp.ind = nan(Nruns, Nq, K); rcpp.dist = nan(Nruns, Nq, K);

tcpp.initruns = zeros(Nruns, 1);
tcpp.queryruns = zeros(Nruns, 1);

for i = 1:Nruns,
    tmark = cputime;
    
    % create tree
    tree = kdtree_build(Xs');
    
    tcpp.initruns(i) = cputime - tmark;
    tmark = cputime;
    
    % query tree
    for iq = 1:Nq,
        [rcpp.ind(i, iq, :), rcpp.dist(i, iq, :)] = kdtree_k_nearest_neighbors(tree, Xq(:, iq), K);
    end;
    
    tcpp.queryruns(i) = cputime - tmark;
    % delete tree (free memory)
    kdtree_delete(tree);    
end;

tcpp.totalruns = tcpp.initruns + tcpp.queryruns;
tcpp.init = mean(tcpp.initruns);
tcpp.query = mean(tcpp.queryruns);
tcpp.total = tcpp.init + tcpp.query;
tcpp.querypersample = tcpp.query / Nq;


%% Test Matlab flat (JIT) implementation -- http://www.mathworks.in/matlabcentral/fileexchange/19345-efficient-k-nearest-neighbor-search-using-jit
tjit = struct;
rjit = struct; rjit.name = 'Matlab JIT'; rjit.ind = nan(Nruns, Nq, K); rjit.dist = nan(Nruns, Nq, K);

tjit.totalruns = zeros(Nruns, 1);

for i = 1:Nruns,
    tmark = cputime;
    % find neighbors
    [rjit.ind(i, :, :), rjit.dist(i, :, :)] = knnsearchx(Xq', Xs', K, S);
    tjit.totalruns(i) = cputime - tmark;
end;

tjit.total = mean(tjit.totalruns);
tjit.totalpersample = tjit.total / Nq;


%% Test Matlab flat, vectorized implementation -- by Lucian Busoniu
tvect = struct;
rvect = struct; rvect.name = 'Matlab flat'; rvect.ind = nan(Nruns, Nq, K); rvect.dist = nan(Nruns, Nq, K);

tvect.totalruns = zeros(Nruns, 1);

for i = 1:Nruns,
    tmark = cputime;
    % find neighbors
    [ind dist] = knn(Xs, Xq, K);
    rvect.ind(i, :, :) = ind';
    rvect.dist(i, :, :) = dist';
    tvect.totalruns(i) = cputime - tmark;
end;

tvect.total = mean(tvect.totalruns);
tvect.totalpersample = tvect.total / Nq;


%% Test Matlab kd-tree implementation -- http://www.mathworks.in/matlabcentral/fileexchange/26649-kdtree-implementation-in-matlab
tkd = struct;
rkd = struct; rkd.name = 'Matlab kd-tree'; rkd.ind = nan(Nruns, Nq, K); rkd.dist = nan(Nruns, Nq, K);

tkd.initruns = zeros(Nruns, 1);
tkd.queryruns = zeros(Nruns, 1);

for i = 1:Nruns,
    tmark = cputime;
    
    % create tree
    tree = kd_buildtree(Xs', 0);    % 0 = no plot
    
    tkd.initruns(i) = cputime - tmark;
    tmark = cputime;
    
    % query tree
    for iq = 1:Nq,
        rkd.ind(i, iq, :) = kd_knn(tree, Xq(:, iq)', K, 0);
    end;
    
    % TODO should also compute distances here, if wishing to compare
    % against other methods
    
    tkd.queryruns(i) = cputime - tmark;
end;

clear tree;

tkd.totalruns = tkd.initruns + tkd.queryruns;
tkd.init = mean(tkd.initruns);
tkd.query = mean(tkd.queryruns);
tkd.total = tkd.init + tkd.query;
tkd.querypersample = tkd.query / Nq;

%% Inspect result for a single method
r = rjit;
graphs = 1;     % try outputting graph if it makes sense
for i = 1:Nruns,
    for iq = 1:Nq,
        ind = r.ind(i, iq, :);
        if graphs && D == 2 && Ns <= 500,
            figure(1); clf;
            scatter(Xs(1, :), Xs(2, :), '.b');
            hold on;
            scatter(Xq(1, iq), Xq(2, iq), 'sk');
            scatter(Xs(1, ind), Xs(2, ind), 'og');
            legend('Database', 'Query point', r.name);
            pause;
        end;
    end;
end;


%% Result comparison between two methods, for identity

r1 = rcpp; r2 = rjit;
% r1 = rcpp; r2 = rvect;
% r1 = rjit; r2 = rvect;
graphs = 1;     % try outputting graph if it makes sense
epsilon = 1e-15;

allvalid = 1;
for i = 1:Nruns,
    for iq = 1:Nq,
        i1 = r1.ind(i, iq, :); d1 = r1.dist(i, iq, :);
        i2 = r2.ind(i, iq, :); d2 = r2.dist(i, iq, :);
        [i1 rearrange] = sort(i1); d1 = d1(rearrange);
        [i2 rearrange] = sort(i2); d2 = d2(rearrange);
        valid = all(i1 == i2) && all(abs(d1 - d2) <= epsilon); 
        allvalid = allvalid && valid;
        if ~valid
            disp(sprintf('Mismatch at i=%d, iq=%d', i, iq));
        end;
        if graphs && D == 2 && Ns <= 500,
            figure(1); clf;
            scatter(Xs(1, :), Xs(2, :), '.b');
            hold on;
            scatter(Xq(1, iq), Xq(2, iq), 'sk');
            scatter(Xs(1, i1), Xs(2, i1), '+r');
            scatter(Xs(1, i2), Xs(2, i2), 'og');
            legend('Database', 'Query point', r1.name, r2.name);
            pause;
        end;
    end;
end;
if allvalid,    disp(sprintf('Test passed: identical results with [%s] and [%s].', r1.name, r2.name));
else            disp(sprintf('*** Test FAILED! Mismatch(es) detected between [%s] and [%s]! ***', r1.name, r2.name)); 
end;

