function varargout = rtlspionline(cfg)
% Real-time, online (semi-)optimistic least-squares policy iteration.
%   THETA = LSPIONLINE(CFG)               - in 'run' mode
%   [HIST, FIGH] = LSPIONLINE(CFG)        - in 'replay' mode
% Real-time, online (semi-)optimistic least-squares policy iteration [1].
% Requires an approximator structure or configuration on the field 'approx', see create_approx for
% details. Can check for convergence (difference between consecutive parameters below eps). If the
% approximator supports so-called "index optimization" (see e.g., the rbfdisc approximator), the
% algorithm will exploit it. Can solve equation A theta = b using mldivide ('inv' mode) or using the
% Sherman-Morisson formula ('recinv' mode). The former is more computationally efficient in Matlab.
%
% The main way of outputting data is a datafile, which will be saved in 'run' and 'resume' modes. 
% The function outputs are provided just for convenience.
%
% Since this is a real-time algorithm, computationally intensive operations should preferably happen
% in-between trials, where the real-time constraints are not so important. E.g., saving data files
% is only done in-between trials, and because solving the linear system is computationally
% intensive, it is also recommended that the interval between policy improvements is a multiple of
% the trial length, so that policy improvements happen after the end of trials.
% 
%
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see str2cfg
% 
% Outputs:
%       THETA       - the computed parameter vector
%       HIST        - the replay history (trajectories)
%       FIGH        - handles to the figures created
%
% Limitation: terminal states (episodic tasks) not supported.
%
% [1] Busoniu, L.; Ernst, D.; De Schutter, B. & Babuska, R. 
%   "Online Least-Squares Policy Iteration for Reinforcement Learning Control"
%   Proceedings 2010 American Control Conference (ACC-10), 2010

% Author: Lucian Busoniu
% Version history:
%
% Detailed execution time statistics are stored in variable "timestat". Fields:
%   init            - scalar, aggregate initialization time (always 0)
%   iter_interact   - time to measure system state, choose action, and apply it
%   iter_computeAb  - vector, time/iter to compute A and b (including the time to compute rewards)
%   iter_qsolve     - vector, time/iter to solve A theta = b
%   iter_peval      - vector, time/iter to perform policy eval (iter_computeAb + qsolve)
%   iter_convtest   - vector, time/iter to test convergence
%   iter_run        - vector, total runtime/iter (iter_interact + iter_peval + iter_convtest)
%   run             - scalar, aggregate runtime (sum of all iter_run)

if nargin < 1, cfg = struct(); end;

% ==== DECLARE CONFIGURATION DEFAULTS ====
CFG.run = 0;                        % run learning
CFG.replay = 0;                     % replay learned policy
CFG.datafile = '';                  % save data to file
CFG.datadir = '';                   % data dir
% main algorithm config
CFG.problem = '';                   % what problem to solve
CFG.system_params = {};             % parameters for problem calling in 'system' mode
CFG.approx = [];                    % approximator object (or config for approx)
CFG.gamma = [];                     % discount factor
CFG.maxtime = [];                   % maximum time to interact with the system; must be multiple of Ts
CFG.trialtime = [];                 % trial length; must be multiple of Ts
CFG.dupd = [];                      % update theta once dupd seconds; preferably multiple of trialtime 
CFG.reset = 'rand';                 % state reset in the beginning of each trial
CFG.method = 'inv';                 % parameter computation method: 'inv', 'recinv'
CFG.invdelta = 0.001;               % use for inverse computation (has to be >0 since LSPI works online)
CFG.recinvdelta = 0.001;            % used for recursive inverse (has to be >0)
CFG.indexoptimized = 1;             % use index-optimized implementation when available (set to 0 to override)
    % note that the index-optimized implementation assumes that the nonzero BFs are arranged in 
    % contiguous, disjoint sequences of the same length (as is the case for discrete-action
    % approximators such as RBFDISC); this flag should be set to 0 when that is not the case
CFG.eps = [];                       % convergence threshold -- empty = do not check (default)
% exploration
CFG.explor = [];                     % start with this exploration rate
CFG.explordecay = [];               % exploration (exponential) decay rate
CFG.decaytype = 'trial';            % decay once per "trial"; or per "second" (for compatibility with lspionline)
CFG.randseed = [];                  % set the random seed (for repeatable results)
% replay config
% uses reset configuration above
CFG.tend = 5;                       % end time for replay
% stats and output config
CFG.dstoretheta = 10;               % save theta at most once each dstoretheta seconds (rarer if theta updates rarer)
CFG.trialsave = 30;                 % save once this many trials
CFG.trialdisp = 5;                  % console feedback at most this many trials
CFG.verb = 3;                       % verbosity: the higher, the more detailed the messages displayed
CFG.silent = 0;                     % work in silence (no console or figure output)
% figure config
CFG = setfigprop(CFG, 'addfields');

% List of fields that define the problem
KEEPFIELDS = {'problem', 'gamma', 'approx'};

% process config
cfg = parseconfig(cfg, CFG);
% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% If running, check presence of data file, warn on possible overwrite
if cfg.run && exist([cfg.datafile '.mat'], 'file'),
    reply = input(['File [' cfg.datafile '.mat] already exists (possible overwrite). Continue? Y/N [N]: '], 's');
    if isempty(reply) || reply == 'N', return; end;
end;

% determine if initialization needed
cfg.init = cfg.run;

% -----------------------------------------------
% Load data if not initializing

if ~cfg.init,        % load data file, making sure that cfg and KEEPFIELDS is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    varnames = who('-file', cfg.datafile);
    % determine which variables are safe to NOT load
    if cfg.replay,
        varnames = rmstring(varnames, 'A', 'b', 'thetah', 'X', 'U', 'R', 'oldtheta');
    end;
    load(cfg.datafile, varnames{:});
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
    approx = revise_approx(approx);     % make sure approx is in latest format
end;

if ~any(strcmp(cfg.method, {'inv', 'recinv'})), 
    error('Unsupported inversion method [%s]', cfg.method); 
end;

% Echo config
dispx('Real-time online LSPI will run with config:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');

% -----------------------------------------------
% Create system and approximator, find dimensionality data

if cfg.init,
    sys = feval(cfg.problem, 'system', cfg.system_params{:});
    Ts = sys.Ts;
    % heuristic check if we already have an approx object
    if isstruct(cfg.approx) && isfield(cfg.approx, 'N') && isfield(cfg.approx, 'phi') ...
            && isfield(cfg.approx, 'q') && isfield(cfg.approx, 'h'),
            % nothing to do
            dispx('Approx object supplied', cfg.verb, 3);
    else    % create approximator using cfg.approx as the config
        cfg.approx = create_approx(sys, cfg.approx);
    end;
    approx = cfg.approx;
    N = approx.N; p = sys.p; q = sys.q;
    if approx.disc,        % initialize action grids and their size
        discU = cfg.approx.U;
        flatU = flat(discU); nU = size(flatU, 2);
        % discUn = zeros(length(discU)); for i=1:length(discU), discUn(i) = length(discU{i}); end;        
    end;
    if ~approx.extendedsyntax,
        error('Approximator that supports extended syntax required');
    end;
end;
% cfg.approx
% cfg.samples

% -----------------------------------------------
% Run algorithm

if cfg.run,
    
    % check interval relationships
    if mod(cfg.maxtime, Ts) || mod(cfg.dupd, Ts), error('maxtime and dupd must be multiples of Ts'); end;
    if mod(cfg.maxtime, cfg.trialtime), error('maxtime must be multiple of trialtime'); end;
    % process various intervals
    Ktrial = cfg.trialtime / Ts;    % number of steps in a trial
    Ksecond = round(1 / Ts);        % approximate(!!) # of steps in a second of real time
    maxtrials = ceil(cfg.maxtime / cfg.trialtime); 
    if ~mod(cfg.dupd, cfg.trialtime),
        updatetrials = cfg.dupd / cfg.trialtime;
        Kupd = false;
    else
        dispx('WARNING! dupd not multiple of trialtime, theta updates may exceed sample period!', cfg.verb, 0);
        Kupd = cfg.dupd / Ts;           % number of steps between theta updates
        updatetrials = false;
    end;

    % maximum possible # of updates of theta (not counting any final update)
    Nupdates = ceil(cfg.maxtime / cfg.dupd);
    
    % check inversion method, init A (according to the method) and b
    recinv = strcmp(cfg.method, 'recinv');          % shorthand
    if recinv,  Ainv = eye(N, N) .* 1/cfg.recinvdelta;
    else        A = eye(N, N); A = A .* cfg.invdelta;
    end;
    b = zeros(N, 1);
    % initialize parameter vector to zeroes; old parameter is the same
    theta = zeros(N, 1); oldtheta = theta;

    % Init histories 
    % variables history (assume fixed-length trials)
    X = nan(p, Ktrial+1, maxtrials);
    U = nan(q, Ktrial+1, maxtrials);
    R = nan(1, Ktrial+1, maxtrials);
    if cfg.dstoretheta > 0,
        % thetah{end} corresponds to the final update of theta
        thetah = {theta};       % for compatibility with old LSPI style
        thetahk = 0;            % theta corresponds to before the 1st sample
    end;
    % first element of this vector will always be NaN, so that elements of these vectors
    % correspond index-wise to the stored values of theta
    % updatesk(end), deltah(end) correspond to the final update of theta
    updatesk = nan(Nupdates+2, 1);  % stores index of sample before update
    deltah = nan(Nupdates+2, 1);
    
    % use index-optimized implementation, if configured and approximator supports it
    cfg.indexoptimized = cfg.indexoptimized && isfield(approx, 'indphi');
    indexoptimized = cfg.indexoptimized;            % shorthand
    if indexoptimized, Nip = approx.Nindphi; end;   % shorthand

    % init counters, indices
    k = 1;               % sample index
    trial = 1;           % index of current trial
    update = 1;          % index of current update
    
    % init execution time, overall and for current update
    timestat.run = 0; 
    timestat.iter_interact(1) = 0; timestat.iter_computeAb(1) = 0;
    % init exploration
    explor = cfg.explor;
        
    % init system
    sys = feval(cfg.problem, 'init', sys);

    % Note there is no sample generation, BF precomputation etc., just
    % variable initialization; hence the init time is taken to be zero
    timestat.init = 0;

    dispx('Performing real-time online LSPI...', cfg.verb, 0);

    if ~isempty(cfg.randseed), rand('twister', cfg.randseed); end;

    % main LSPI-online loop
    conv = 0;
    tic;    % start real-time counter
    while trial <= maxtrials && ~conv,
        % initialize trial
        % reset to initial state
        X(:, 1, trial) = feval(cfg.problem, 'reset', sys, cfg.reset);
        % step counter, terminal flag
        ktrial = 1; % term = 0;

        % dispx(sprintf('===> Trial %d started // k=%d ktrial=%d trial=%d t=%f', trial, k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);
        
        % note we iterate one extra step, to allow the update for the last transition 
        % (to sample # Ktrial+1) to run
        while ktrial <= Ktrial+1, % && ~term,        
            
            tmark = cputime;
            % dispx(sprintf('===> Measuring x // k=%d ktrial=%d trial=%d t=%f', k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);
            % measure state (unless first step in trial, in which case the 'reset' mode already
            % returned it
            if ktrial > 1,
                X(:, ktrial, trial) = feval(cfg.problem, 'readx', sys);
            end;

            % === BEGIN TIME-CRITICAL -- action choice code
            if ktrial <= Ktrial,      % must not apply action after last sample
                % choose action according to current state
                exploring = rand < explor;
                % choose control action
                if exploring,  % explore, choose uniform random action
                    if approx.disc,    % flatU is a set of discrete actions, each action on a column
                        ucandidate = flatU(:, unidrnd(nU));
                    else        % U is continuous, assumed symmetrical w/ bounds in maxu
                        ucandidate = (2*rand(q, 1) - 1) .* sys.maxu;
                    end;
                else            % exploit using policy greedy in the current theta
                    ucandidate = approx.h(approx, theta, X(:, ktrial, trial));
                end;
                % dispx(sprintf('===> Applying u // k=%d ktrial=%d trial=%d t=%f', k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);
                % apply the action
                % note the system may return a modified (e.g., filtered) action
                % if this happens, discrete-action approximators may be broken!
                U(:, ktrial, trial) = feval(cfg.problem, 'writeu', sys, ucandidate);
            else
                % trial has ended, make the system idle
                feval(cfg.problem, 'idle', sys);
            end;
            % === END TIME-CRITICAL
            timestat.iter_interact(update) = timestat.iter_interact(update) + cputime - tmark;

            % from now on, we have the whole sample time -- but we should not exceed it
            
            % if we have a transition to work on, perform a learning update
            %               -- always working on the previous transition sample!
            if ktrial > 1,
                % dispx(sprintf('===> Processing transition %d -> %d // k=%d ktrial=%d trial=%d t=%f', ktrial-1, ktrial, k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);
                tmark = cputime;
                % apply system function to compute reward & check if transition was terminal
                [R(1, ktrial, trial), term] = ...
                    feval(sys.fun, sys, X(:, ktrial-1, trial), U(:, ktrial-1, trial), X(:, ktrial, trial));
                if term, error('Terminal state support not implemented'); end;

                % if we are not exploring, and not at the last step, then we already chose a greedy
                % action for the current (last) state; otherwise, we need to do it now
                if ~exploring && ktrial <= Ktrial, 
                    up = ucandidate;
                else
                    up = approx.h(approx, theta, X(:, ktrial, trial));
                end;
                
                % update A and b
                if indexoptimized,
                    % compute indices and values of phi(x, u) and phi(x', h(x'))
                    [off, phixu] = approx.indphi(approx, X(:, ktrial-1, trial), U(:, ktrial-1, trial));
                    [offp, phixpup] = approx.indphi(approx, X(:, ktrial, trial), up);
                    ustart = off+1; uend = off+Nip; vstart = offp+1; vend = offp+Nip;
                    % (Note here we assume that ranges in BF vectors do NOT overlap and have EQUAL sizes)
                    % update A and b, using the indices
                    if recinv,
                        % 1) update the block on the diagonal corresponding to (x, u) X (x, u)
                        num = Ainv(:, ustart:uend) * (phixu * phixu') * Ainv(ustart:uend, :);
                        den = 1 + phixu' * Ainv(ustart:uend, ustart:uend) * phixu;
                        Ainv = Ainv - num ./ den;
                        % 2) update corresponding to off-diagonal (x, u) X (x', h(x'))
                        num = Ainv(:, ustart:uend) * (phixu * phixpup') * Ainv(vstart:vend, :);
                        den = 1 - cfg.gamma * (phixpup' * Ainv(vstart:vend, ustart:uend) * phixu);
                        Ainv = Ainv + num .* (cfg.gamma/den);
                    else
                        A(ustart:uend, ustart:uend) = A(ustart:uend, ustart:uend) + phixu * phixu';
                        A(ustart:uend, vstart:vend) = A(ustart:uend, vstart:vend) - cfg.gamma * phixu * phixpup';
                    end;
                    b(ustart:uend) = b(ustart:uend) + phixu * R(:, ktrial, trial);
                else
                    % compute phi(x, u) and phi(x', h(x'))
                    phixu = approx.phi(approx, X(:, ktrial-1, trial), U(:, ktrial-1, trial));
                    phixpup = approx.phi(approx, X(:, ktrial, trial), up);
                    % update A and b
                    if recinv,
                        if k == 1, 
                            dispx('WARNING! Using recinv update with full basis vectors: will run slowly.', cfg.verb, 0); 
                        end;
                        v = phixu - cfg.gamma * phixpup;
                        Ainv = Ainv - (Ainv * phixu * v' * Ainv) / (1 + v' * Ainv * phixu);
                    else
                        A = A + phixu * (phixu - cfg.gamma * phixpup)';
                    end;
                    b = b + phixu * R(:, ktrial, trial);
                end;
                timestat.iter_computeAb(update) = timestat.iter_computeAb(update) + cputime - tmark;
            end;
            
            % policy evaluation and improvement, either:
            % 1. synchronized with trial ends (first test), or:
            % 2. once each Kupd transitions (second test); because the kth transition is only
            % processed at step k+1, we must delay the test by 1 step, hence mod(k-1, Kupd)
            % Note (for case 2) we do NOT update right at the first sample in each trial, because:
            %      - in trial 1, no samples have been processed
            %      - in trial n > 1, k had the same value when processing the final transition of
            %      trial n-1, hence the update was trigerred then (it is better to have it trigerred
            %      then than during 1st sample, as the time constraints are relaxed at the end of
            %      trial)
            if (updatetrials && ktrial == Ktrial+1 && ~mod(trial, updatetrials)) ...   
                    || (Kupd && ktrial > 1 && ~mod(k - 1, Kupd)),
                % dispx(sprintf('===> Policy improvement %d // k=%d ktrial=%d trial=%d t=%f', update, k, ktrial, trial, (k-1)*Ts), cfg.verb, 5);
                tmark = cputime;
                % remember after which transition we update
                updatesk(update) = k - 1;
                % note normalization is also by k-1, the actual # of processed transitions
                if recinv,  theta = (Ainv.*(k-1)) * (b./(k-1));
                else        theta = (A./(k-1)) \ (b./(k-1));
                end;
                timestat.iter_qsolve(update) = cputime - tmark; tmark = cputime;
                deltah(update+1) = norm(theta - oldtheta, 2);
                oldtheta = theta;
                % check convergence condition, if configured
                if ~isempty(cfg.eps), conv = deltah(update+1) <= cfg.eps; end;
                timestat.iter_convtest(update) = cputime - tmark; % below not counted
                % store new parameter vector if required and passed time boundary
                if cfg.dstoretheta > 0,
                    if (update == 1 && floor(k*Ts / cfg.dstoretheta) > 0) ...
                        || (update > 1 && floor(k*Ts / cfg.dstoretheta) > floor(updatesk(update-1)*Ts / cfg.dstoretheta)),
                        % save a theta snapshot
                        thetah{end+1} = theta; 
                        % remember after which transition we took the snapshot
                        thetahk(end+1) = k - 1;
                    end;
                end;
 
                % update aggregate time stats
                timestat.iter_peval(update) = timestat.iter_computeAb(update) + timestat.iter_qsolve(update);
                timestat.iter_run(update) = timestat.iter_interact(update) + timestat.iter_peval(update) + timestat.iter_convtest(update);
                timestat.run = timestat.run + timestat.iter_run(update);
                
                dispx(sprintf('Updated %d times, t=%.3f', update, (k-1)*Ts), cfg.verb, 3);

                update = update + 1;            % increment updates counter
                % initialize per-sample time stats
                timestat.iter_interact(update) = 0; timestat.iter_computeAb(update) = 0;
            end;
            
            % if decaytype = second, decay exploration (approximately) once every second
            if ~mod(k, Ksecond) && cfg.decaytype(1) == 's', explor = explor * cfg.explordecay; end;

            % increment step counter within trial
            ktrial = ktrial + 1;
            
            % the step is done; unless we are taking the final extra step:
            if ktrial <= Ktrial+1,             
                % increment overall sample counter
                k = k + 1; 
                % wait out the rest of the sample period
                while toc < sys.Ts, end; tic;
            end;

        end;    % steps loop
        
        % if decaytype = trial, decay exploration once every trial
        if cfg.decaytype(1) == 't', explor = explor * cfg.explordecay; end;
        
        % console feedback, data backup
        if ~mod(trial, cfg.trialdisp),
            dispx(sprintf('Trial #%d done, updates=%d, delta=%f', trial, update-1, deltah(update)), cfg.verb, 2);
        end;
        if ~mod(trial, cfg.trialsave),
            save(cfg.datafile);
            dispx(sprintf('Data after trial #%d saved to [%s].', trial, cfg.datafile), cfg.verb, 1);
        end;
        
        trial = trial + 1;
    end;        % trials loop

    % NOTE difference from the simulation lspionline
    % lspionline performs a final update here, whereas:
    % if maxtime multiple of dupd, then we do the same update at the end of last trial 
    % otherwise, we do not, so we are one update short of lspionline, and ignore the samples after
    % the last update

    % close system
    sys = feval(sys.problem, 'close', sys);    

    if conv,        dispx('Convergence detected. Algorithm stopped.', cfg.verb, 0);
    else            dispx(['maxtrials=' num2str(maxtrials) ' exhausted. Algorithm stopped'], cfg.verb, 0);
    end;
    
    varargout = {theta};
end;


% Options below might create figures, init figure handles array
figh = [];

% Replay using learned policy
if cfg.replay,

    % init system
    sys = feval(cfg.problem, 'init', sys);        
    % reset to initial state
    x0 = feval(cfg.problem, 'reset', sys, cfg.reset);
    dispx(['Controlling from x0=' num2str(reshape(x0, 1, [])) ], cfg.verb, 0);

    % init history
    t = 0 : sys.Ts : cfg.tend;
    Ns = length(t)-1;       % number of samples / time instances at which control is applied
    x = zeros(sys.p, length(t)); x(:, 1) = x0;
    u = zeros(sys.q, length(t)); u(:, end) = NaN;
    r = zeros(1, length(t)); r(1) = NaN;

    % steps loop
    tic;
    for k = 1:Ns+1,
        x(:, k) = feval(cfg.problem, 'readx', sys);
        % time-critical section
        if k <= Ns,     % need to apply control
            ucandidate = approx.h(approx, theta, x(:, k));
            u(:, k) = feval(cfg.problem, 'writeu', sys, ucandidate);
        end;
        % now we have the rest of the sample time
        if k > 1,       % we have a transition to process
            [r(1, k), term] = feval(sys.fun, sys, x(:, k-1), u(:, k-1), x(:, k));
        end;
        if term, error('Terminal state support not implemented'); end;
        while toc < Ts, end; tic;
    end;
    % close system
    sys = feval(cfg.problem, 'close', sys);        

    % plot history & optionally save figures
    history.t = t(1:Ns+1); history.x = x(:, 1:Ns+1); history.u = u(:, 1:Ns+1); history.r = r(1:Ns+1);
    history.R = discreturn(cfg, history.r, Ns, term);
    if ~cfg.silent,
        if isfield(sys, 'plotfun'),
            mpfigh = feval(model.plotfun, history);
            if ~isempty(mpfigh), figh = mpfigh; end;
        else
            figh = plothistory(history);
        end;
        setfigprop(cfg);
        % save if requested
        saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
    end;
end;        % IF replay


% -----------------------------------------------
% Backup data

if cfg.run && ~isempty(cfg.datafile),
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir;
        if (datadir(end) == '\'), datadir = datadir(1:end-1); end;
    end;
    cfg.datafile = [datadir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;



end
% lspi() RETURNING varargout =================================================================
